################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Union

from pyflink.java_gateway import get_gateway
from pyflink.table import ExplainDetail
from pyflink.table.table_descriptor import TableDescriptor
from pyflink.table.table_result import TableResult
from pyflink.util.java_utils import to_j_explain_detail_arr

__all__ = ['StatementSet']


class StatementSet(object):
    """
    A :class:`~StatementSet` accepts pipelines defined by DML statements or :class:`~Table` objects.
    The planner can optimize all added statements together and then submit them as one job.

    The added statements will be cleared when calling the :func:`~StatementSet.execute` method.

    .. versionadded:: 1.11.0
    """

    def __init__(self, _j_statement_set, t_env):
        self._j_statement_set = _j_statement_set
        self._t_env = t_env

    def add_insert_sql(self, stmt: str) -> 'StatementSet':
        """
        add insert statement to the set.

        :param stmt: The statement to be added.
        :return: current StatementSet instance.

        .. versionadded:: 1.11.0
        """
        self._j_statement_set.addInsertSql(stmt)
        return self

    def attach_as_datastream(self):
        """
        Optimizes all statements as one entity and adds them as transformations to the underlying
        StreamExecutionEnvironment.

        Use :func:`~pyflink.datastream.StreamExecutionEnvironment.execute` to execute them.

        The added statements will be cleared after calling this method.

        .. versionadded:: 1.16.0
        """
        self._j_statement_set.attachAsDataStream()

    def add_insert(self,
                   target_path_or_descriptor: Union[str, TableDescriptor],
                   table,
                   overwrite: bool = False) -> 'StatementSet':
        """
        Adds a statement that the pipeline defined by the given Table object should be written to a
        table (backed by a DynamicTableSink) that was registered under the specified path or
        expressed via the given TableDescriptor.

        1. When target_path_or_descriptor is a tale path:

            See the documentation of :func:`~TableEnvironment.use_database` or
            :func:`~TableEnvironment.use_catalog` for the rules on the path resolution.

        2. When target_path_or_descriptor is a table descriptor:

            The given TableDescriptor is registered as an inline (i.e. anonymous) temporary catalog
            table (see :func:`~TableEnvironment.create_temporary_table`).

            Then a statement is added to the statement set that inserts the Table object's pipeline
            into that temporary table.

            This method allows to declare a Schema for the sink descriptor. The declaration is
            similar to a {@code CREATE TABLE} DDL in SQL and allows to:

                1. overwrite automatically derived columns with a custom DataType
                2. add metadata columns next to the physical columns
                3. declare a primary key

            It is possible to declare a schema without physical/regular columns. In this case, those
            columns will be automatically derived and implicitly put at the beginning of the schema
            declaration.

            Examples:
            ::

                >>> stmt_set = table_env.create_statement_set()
                >>> source_table = table_env.from_path("SourceTable")
                >>> sink_descriptor = TableDescriptor.for_connector("blackhole") \\
                ...     .schema(Schema.new_builder()
                ...         .build()) \\
                ...     .build()
                >>> stmt_set.add_insert(sink_descriptor, source_table)

            .. note:: add_insert for a table descriptor (case 2.) was added from
                flink 1.14.0.

        :param target_path_or_descriptor: The path of the registered
            :class:`~pyflink.table.TableSink` or the descriptor describing the sink table into which
            data should be inserted to which the :class:`~pyflink.table.Table` is written.
        :param table: The Table to add.
        :type table: pyflink.table.Table
        :param overwrite: Indicates whether the insert should overwrite existing data or not.
        :return: current StatementSet instance.

        .. versionadded:: 1.11.0
        """
        if isinstance(target_path_or_descriptor, str):
            self._j_statement_set.addInsert(target_path_or_descriptor, table._j_table, overwrite)
        else:
            self._j_statement_set.addInsert(
                target_path_or_descriptor._j_table_descriptor, table._j_table, overwrite)
        return self

    def explain(self, *extra_details: ExplainDetail) -> str:
        """
        returns the AST and the execution plan of all statements and Tables.

        :param extra_details: The extra explain details which the explain result should include,
                              e.g. estimated cost, changelog mode for streaming
        :return: All statements and Tables for which the AST and execution plan will be returned.

        .. versionadded:: 1.11.0
        """
        TEXT = get_gateway().jvm.org.apache.flink.table.api.ExplainFormat.TEXT
        j_extra_details = to_j_explain_detail_arr(extra_details)
        return self._j_statement_set.explain(TEXT, j_extra_details)

    def execute(self) -> TableResult:
        """
        execute all statements and Tables as a batch.

        .. note::
            The added statements and Tables will be cleared when executing this method.

        :return: execution result.

        .. versionadded:: 1.11.0
        """
        self._t_env._before_execute()
        return TableResult(self._j_statement_set.execute())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import json
import logging
import sys

from pyflink.common import Types
from pyflink.datastream import StreamExecutionEnvironment


def show(ds, env):
    ds.print()
    env.execute()


def basic_operations():
    env = StreamExecutionEnvironment.get_execution_environment()
    env.set_parallelism(1)

    # define the source
    ds = env.from_collection(
        collection=[
            (1, '{"name": "Flink", "tel": 123, "addr": {"country": "Germany", "city": "Berlin"}}'),
            (2, '{"name": "hello", "tel": 135, "addr": {"country": "China", "city": "Shanghai"}}'),
            (3, '{"name": "world", "tel": 124, "addr": {"country": "USA", "city": "NewYork"}}'),
            (4, '{"name": "PyFlink", "tel": 32, "addr": {"country": "China", "city": "Hangzhou"}}')
        ],
        type_info=Types.ROW_NAMED(["id", "info"], [Types.INT(), Types.STRING()])
    )

    # map
    def update_tel(data):
        # parse the json
        json_data = json.loads(data.info)
        json_data['tel'] += 1
        return data.id, json.dumps(json_data)

    show(ds.map(update_tel), env)
    # (1, '{"name": "Flink", "tel": 124, "addr": {"country": "Germany", "city": "Berlin"}}')
    # (2, '{"name": "hello", "tel": 136, "addr": {"country": "China", "city": "Shanghai"}}')
    # (3, '{"name": "world", "tel": 125, "addr": {"country": "USA", "city": "NewYork"}}')
    # (4, '{"name": "PyFlink", "tel": 33, "addr": {"country": "China", "city": "Hangzhou"}}')

    # filter
    show(ds.filter(lambda data: data.id == 1).map(update_tel), env)
    # (1, '{"name": "Flink", "tel": 124, "addr": {"country": "Germany", "city": "Berlin"}}')

    # key by
    show(ds.map(lambda data: (json.loads(data.info)['addr']['country'],
                              json.loads(data.info)['tel']))
           .key_by(lambda data: data[0]).sum(1), env)
    # ('Germany', 123)
    # ('China', 135)
    # ('USA', 124)
    # ('China', 167)


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    basic_operations()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import json
import logging
import sys

from pyflink.datastream import StreamExecutionEnvironment


def process_json_data():
    env = StreamExecutionEnvironment.get_execution_environment()

    # define the source
    ds = env.from_collection(
        collection=[
            (1, '{"name": "Flink", "tel": 123, "addr": {"country": "Germany", "city": "Berlin"}}'),
            (2, '{"name": "hello", "tel": 135, "addr": {"country": "China", "city": "Shanghai"}}'),
            (3, '{"name": "world", "tel": 124, "addr": {"country": "USA", "city": "NewYork"}}'),
            (4, '{"name": "PyFlink", "tel": 32, "addr": {"country": "China", "city": "Hangzhou"}}')]
    )

    def update_tel(data):
        # parse the json
        json_data = json.loads(data[1])
        json_data['tel'] += 1
        return data[0], json_data

    def filter_by_country(data):
        # the json data could be accessed directly, there is no need to parse it again using
        # json.loads
        return "China" in data[1]['addr']['country']

    ds.map(update_tel).filter(filter_by_country).print()

    # submit for execution
    env.execute()


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    process_json_data()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import sys

import argparse
from typing import Iterable

from pyflink.datastream.connectors.file_system import FileSink, OutputFileConfig, RollingPolicy

from pyflink.common import Types, WatermarkStrategy, Time, Encoder
from pyflink.common.watermark_strategy import TimestampAssigner
from pyflink.datastream import StreamExecutionEnvironment, ProcessWindowFunction
from pyflink.datastream.window import SlidingEventTimeWindows, TimeWindow


class MyTimestampAssigner(TimestampAssigner):
    def extract_timestamp(self, value, record_timestamp) -> int:
        return int(value[1])


class CountWindowProcessFunction(ProcessWindowFunction[tuple, tuple, str, TimeWindow]):
    def process(self,
                key: str,
                context: ProcessWindowFunction.Context[TimeWindow],
                elements: Iterable[tuple]) -> Iterable[tuple]:
        return [(key, context.window().start, context.window().end, len([e for e in elements]))]


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--output',
        dest='output',
        required=False,
        help='Output file to write results to.')

    argv = sys.argv[1:]
    known_args, _ = parser.parse_known_args(argv)
    output_path = known_args.output

    env = StreamExecutionEnvironment.get_execution_environment()
    # write all the data to one file
    env.set_parallelism(1)

    # define the source
    data_stream = env.from_collection([
        ('hi', 1), ('hi', 2), ('hi', 3), ('hi', 4), ('hi', 5), ('hi', 8), ('hi', 9), ('hi', 15)],
        type_info=Types.TUPLE([Types.STRING(), Types.INT()]))

    # define the watermark strategy
    watermark_strategy = WatermarkStrategy.for_monotonous_timestamps() \
        .with_timestamp_assigner(MyTimestampAssigner())

    ds = data_stream.assign_timestamps_and_watermarks(watermark_strategy) \
        .key_by(lambda x: x[0], key_type=Types.STRING()) \
        .window(SlidingEventTimeWindows.of(Time.milliseconds(5), Time.milliseconds(2))) \
        .process(CountWindowProcessFunction(),
                 Types.TUPLE([Types.STRING(), Types.INT(), Types.INT(), Types.INT()]))

    # define the sink
    if output_path is not None:
        ds.sink_to(
            sink=FileSink.for_row_format(
                base_path=output_path,
                encoder=Encoder.simple_string_encoder())
            .with_output_file_config(
                OutputFileConfig.builder()
                .with_part_prefix("prefix")
                .with_part_suffix(".ext")
                .build())
            .with_rolling_policy(RollingPolicy.default_rolling_policy())
            .build()
        )
    else:
        print("Printing result to stdout. Use --output to specify output path.")
        ds.print()

    # submit for execution
    env.execute()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.java_gateway import get_gateway

__all__ = ['ResultKind']


class ResultKind(object):
    """
    ResultKind defines the types of the result.

    :data:`SUCCESS`:

    The statement (e.g. DDL, USE) executes successfully, and the result only contains a simple "OK".

    :data:`SUCCESS_WITH_CONTENT`:

    The statement (e.g. DML, DQL, SHOW) executes successfully, and the result contains important
    content.

    .. versionadded:: 1.11.0
    """

    SUCCESS = 0
    SUCCESS_WITH_CONTENT = 1

    @staticmethod
    def _from_j_result_kind(j_result_kind):
        gateway = get_gateway()
        JResultKind = gateway.jvm.org.apache.flink.table.api.ResultKind
        if j_result_kind == JResultKind.SUCCESS:
            return ResultKind.SUCCESS
        elif j_result_kind == JResultKind.SUCCESS_WITH_CONTENT:
            return ResultKind.SUCCESS_WITH_CONTENT
        else:
            raise Exception("Unsupported Java result kind: %s" % j_result_kind)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common import SerializationSchema, TypeInformation, typeinfo, DeserializationSchema
from pyflink.java_gateway import get_gateway


__all__ = [
    'JsonRowDeserializationSchema',
    'JsonRowSerializationSchema'
]


class JsonRowDeserializationSchema(DeserializationSchema):
    """
    Deserialization schema from JSON to Flink types.

    Deserializes a byte[] message as a JSON object and reads the specified fields.

    Failures during deserialization are forwarded as wrapped IOExceptions.
    """
    def __init__(self, j_deserialization_schema):
        super(JsonRowDeserializationSchema, self).__init__(j_deserialization_schema)

    @staticmethod
    def builder():
        """
        A static method to get a Builder for JsonRowDeserializationSchema.
        """
        return JsonRowDeserializationSchema.Builder()

    class Builder(object):
        """
        Builder for JsonRowDeserializationSchema.
        """

        def __init__(self):
            self._type_info = None
            self._fail_on_missing_field = False
            self._ignore_parse_errors = False

        def type_info(self, type_info: TypeInformation):
            """
            Creates a JSON deserialization schema for the given type information.

            :param type_info: Type information describing the result type. The field names of Row
                              are used to parse the JSON properties.
            """
            self._type_info = type_info
            return self

        def json_schema(self, json_schema: str):
            """
            Creates a JSON deserialization schema for the given JSON schema.

            :param json_schema: JSON schema describing the result type.
            """
            if json_schema is None:
                raise TypeError("The json_schema must not be None.")
            j_type_info = get_gateway().jvm \
                .org.apache.flink.formats.json.JsonRowSchemaConverter.convert(json_schema)
            self._type_info = typeinfo._from_java_type(j_type_info)
            return self

        def fail_on_missing_field(self):
            """
            Configures schema to fail if a JSON field is missing. A missing field is ignored and the
            field is set to null by default.
            """
            self._fail_on_missing_field = True
            return self

        def ignore_parse_errors(self):
            """
            Configures schema to fail when parsing json failed. An exception will be thrown when
            parsing json fails.
            """
            self._ignore_parse_errors = True
            return self

        def build(self):
            JBuilder = get_gateway().jvm.org.apache.flink.formats.json.JsonRowDeserializationSchema\
                .Builder
            j_builder = JBuilder(self._type_info.get_java_type_info())

            if self._fail_on_missing_field:
                j_builder = j_builder.failOnMissingField()

            if self._ignore_parse_errors:
                j_builder = j_builder.ignoreParseErrors()

            j_deserialization_schema = j_builder.build()
            return JsonRowDeserializationSchema(j_deserialization_schema=j_deserialization_schema)


class JsonRowSerializationSchema(SerializationSchema):
    """
    Serialization schema that serializes an object of Flink types into a JSON bytes. Serializes the
    input Flink object into a JSON string and converts it into byte[].

    Result byte[] message can be deserialized using JsonRowDeserializationSchema.
    """

    def __init__(self, j_serialization_schema):
        super(JsonRowSerializationSchema, self).__init__(j_serialization_schema)

    @staticmethod
    def builder():
        return JsonRowSerializationSchema.Builder()

    class Builder(object):
        """
        Builder for JsonRowSerializationSchema.
        """
        def __init__(self):
            self._type_info = None

        def with_type_info(self, type_info: TypeInformation):
            """
            Creates a JSON serialization schema for the given type information.

            :param type_info: Type information describing the result type. The field names of Row
                              are used to parse the JSON properties.
            """
            self._type_info = type_info
            return self

        def build(self):
            if self._type_info is None:
                raise TypeError("Typeinfo should be set.")

            j_builder = get_gateway().jvm \
                .org.apache.flink.formats.json.JsonRowSerializationSchema.builder()

            j_schema = j_builder.withTypeInfo(self._type_info.get_java_type_info()).build()
            return JsonRowSerializationSchema(j_serialization_schema=j_schema)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import sys
from abc import ABCMeta
from collections import OrderedDict

from py4j.java_gateway import get_method
from typing import Dict, Union

from pyflink.java_gateway import get_gateway
from pyflink.table.table_schema import TableSchema
from pyflink.table.types import DataType, _to_java_data_type

__all__ = [
    'Rowtime',
    'Schema'
]


class Descriptor(object, metaclass=ABCMeta):
    """
    Base class of the descriptors that adds a set of string-based, normalized properties for
    describing DDL information.

    Typical characteristics of a descriptor are:
    - descriptors have a default constructor
    - descriptors themselves contain very little logic
    - corresponding validators validate the correctness (goal: have a single point of validation)

    A descriptor is similar to a builder in a builder pattern, thus, mutable for building
    properties.
    """

    def __init__(self, j_descriptor):
        self._j_descriptor = j_descriptor

    def to_properties(self) -> Dict:
        """
        Converts this descriptor into a dict of properties.

        :return: Dict object contains all of current properties.
        """
        return dict(self._j_descriptor.toProperties())


class Rowtime(Descriptor):
    """
    Rowtime descriptor for describing an event time attribute in the schema.
    """

    def __init__(self):
        gateway = get_gateway()
        self._j_rowtime = gateway.jvm.Rowtime()
        super(Rowtime, self).__init__(self._j_rowtime)

    def timestamps_from_field(self, field_name: str):
        """
        Sets a built-in timestamp extractor that converts an existing LONG or TIMESTAMP field into
        the rowtime attribute.

        :param field_name: The field to convert into a rowtime attribute.
        :return: This rowtime descriptor.
        """
        self._j_rowtime = self._j_rowtime.timestampsFromField(field_name)
        return self

    def timestamps_from_source(self) -> 'Rowtime':
        """
        Sets a built-in timestamp extractor that converts the assigned timestamps from a DataStream
        API record into the rowtime attribute and thus preserves the assigned timestamps from the
        source.

        .. note::

            This extractor only works in streaming environments.

        :return: This rowtime descriptor.
        """
        self._j_rowtime = self._j_rowtime.timestampsFromSource()
        return self

    def timestamps_from_extractor(self, extractor: str) -> 'Rowtime':
        """
        Sets a custom timestamp extractor to be used for the rowtime attribute.

        :param extractor: The java fully-qualified class name of the TimestampExtractor to extract
                          the rowtime attribute from the physical type. The TimestampExtractor must
                          have a public no-argument constructor and can be founded by
                          in current Java classloader.
        :return: This rowtime descriptor.
        """
        gateway = get_gateway()
        self._j_rowtime = self._j_rowtime.timestampsFromExtractor(
            gateway.jvm.Thread.currentThread().getContextClassLoader().loadClass(extractor)
                   .newInstance())
        return self

    def watermarks_periodic_ascending(self) -> 'Rowtime':
        """
        Sets a built-in watermark strategy for ascending rowtime attributes.

        Emits a watermark of the maximum observed timestamp so far minus 1. Rows that have a
        timestamp equal to the max timestamp are not late.

        :return: This rowtime descriptor.
        """
        self._j_rowtime = self._j_rowtime.watermarksPeriodicAscending()
        return self

    def watermarks_periodic_bounded(self, delay: int) -> 'Rowtime':
        """
        Sets a built-in watermark strategy for rowtime attributes which are out-of-order by a
        bounded time interval.

        Emits watermarks which are the maximum observed timestamp minus the specified delay.

        :param delay: Delay in milliseconds.
        :return: This rowtime descriptor.
        """
        self._j_rowtime = self._j_rowtime.watermarksPeriodicBounded(delay)
        return self

    def watermarks_from_source(self) -> 'Rowtime':
        """
        Sets a built-in watermark strategy which indicates the watermarks should be preserved from
        the underlying DataStream API and thus preserves the assigned watermarks from the source.

        :return: This rowtime descriptor.
        """
        self._j_rowtime = self._j_rowtime.watermarksFromSource()
        return self

    def watermarks_from_strategy(self, strategy: str) -> 'Rowtime':
        """
        Sets a custom watermark strategy to be used for the rowtime attribute.

        :param strategy: The java fully-qualified class name of the WatermarkStrategy. The
                         WatermarkStrategy must have a public no-argument constructor and can be
                         founded by in current Java classloader.
        :return: This rowtime descriptor.
        """
        gateway = get_gateway()
        self._j_rowtime = self._j_rowtime.watermarksFromStrategy(
            gateway.jvm.Thread.currentThread().getContextClassLoader().loadClass(strategy)
                   .newInstance())
        return self


class Schema(Descriptor):
    """
    Describes a schema of a table.

    .. note::

        Field names are matched by the exact name by default (case sensitive).
    """

    def __init__(self, schema=None, fields=None, rowtime=None):
        """
        Constructor of Schema descriptor.

        :param schema: The :class:`TableSchema` object.
        :param fields: Dict of fields with the field name and the data type or type string stored.
        :param rowtime: A :class:`RowTime` that Specifies the previously defined field as an
                        event-time attribute.
        """
        gateway = get_gateway()
        self._j_schema = gateway.jvm.org.apache.flink.table.legacy.descriptors.Schema()
        super(Schema, self).__init__(self._j_schema)

        if schema is not None:
            self.schema(schema)

        if fields is not None:
            self.fields(fields)

        if rowtime is not None:
            self.rowtime(rowtime)

    def schema(self, table_schema: 'TableSchema') -> 'Schema':
        """
        Sets the schema with field names and the types. Required.

        This method overwrites existing fields added with
        :func:`~pyflink.table.descriptors.Schema.field`.

        :param table_schema: The :class:`TableSchema` object.
        :return: This schema object.
        """
        self._j_schema = self._j_schema.schema(table_schema._j_table_schema)
        return self

    def field(self, field_name: str, field_type: Union[DataType, str]) -> 'Schema':
        """
        Adds a field with the field name and the data type or type string. Required.
        This method can be called multiple times. The call order of this method defines
        also the order of the fields in a row. Here is a document that introduces the type strings:
        https://nightlies.apache.org/flink/flink-docs-stable/dev/table/connect.html#type-strings

        :param field_name: The field name.
        :param field_type: The data type or type string of the field.
        :return: This schema object.
        """
        if isinstance(field_type, str):
            self._j_schema = self._j_schema.field(field_name, field_type)
        else:
            self._j_schema = self._j_schema.field(field_name, _to_java_data_type(field_type))
        return self

    def fields(self, fields: Dict[str, Union[DataType, str]]) -> 'Schema':
        """
        Adds a set of fields with the field name and the data type or type string stored in a
        list.

        :param fields: Dict of fields with the field name and the data type or type string
                       stored.
                       E.g, [('int_field', DataTypes.INT()), ('string_field', DataTypes.STRING())].
        :return: This schema object.

        .. versionadded:: 1.11.0
        """
        if sys.version_info[:2] <= (3, 5) and not isinstance(fields, OrderedDict):
            raise TypeError("Must use OrderedDict type in python3.5 or older version to key the "
                            "schema in insert order.")
        elif sys.version_info[:2] > (3, 5) and not isinstance(fields, (OrderedDict, dict)):
            raise TypeError("fields must be stored in a dict or OrderedDict")

        for field_name, field_type in fields.items():
            self.field(field_name=field_name, field_type=field_type)
        return self

    def from_origin_field(self, origin_field_name: str) -> 'Schema':
        """
        Specifies the origin of the previously defined field. The origin field is defined by a
        connector or format.

        E.g. field("myString", Types.STRING).from_origin_field("CSV_MY_STRING")

        .. note::

            Field names are matched by the exact name by default (case sensitive).

        :param origin_field_name: The origin field name.
        :return: This schema object.
        """
        self._j_schema = get_method(self._j_schema, "from")(origin_field_name)
        return self

    def proctime(self) -> 'Schema':
        """
        Specifies the previously defined field as a processing-time attribute.

        E.g. field("proctime", Types.SQL_TIMESTAMP_LTZ).proctime()

        :return: This schema object.
        """
        self._j_schema = self._j_schema.proctime()
        return self

    def rowtime(self, rowtime: Rowtime) -> 'Schema':
        """
        Specifies the previously defined field as an event-time attribute.

        E.g. field("rowtime", Types.SQL_TIMESTAMP).rowtime(...)

        :param rowtime: A :class:`RowTime`.
        :return: This schema object.
        """
        self._j_schema = self._j_schema.rowtime(rowtime._j_rowtime)
        return self



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import abc
from enum import Enum
from typing import List, Union

from pyflink.datastream.connectors import Sink, DeliveryGuarantee
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import to_jarray


__all__ = ['FlushBackoffType',
           'ElasticsearchEmitter',
           'Elasticsearch6SinkBuilder',
           'Elasticsearch7SinkBuilder',
           'ElasticsearchSink']


class FlushBackoffType(Enum):
    """
    Used to control whether the sink should retry failed requests at all or with which kind back off
    strategy.

    :data: `CONSTANT`:

    After every failure, it waits a configured time until the retries are exhausted.

    :data: `EXPONENTIAL`:

    After every failure, it waits initially the configured time and increases the waiting time
    exponentially until the retries are exhausted.

    :data: `NONE`:

    The failure is not retried.
    """

    CONSTANT = 0,
    EXPONENTIAL = 1,
    NONE = 2,

    def _to_j_flush_backoff_type(self):
        JFlushBackoffType = get_gateway().jvm \
            .org.apache.flink.connector.elasticsearch.sink.FlushBackoffType
        return getattr(JFlushBackoffType, self.name)


class ElasticsearchEmitter(object):
    """
    Emitter which is used by sinks to prepare elements for sending them to Elasticsearch.
    """

    def __init__(self, j_emitter):
        self._j_emitter = j_emitter

    @staticmethod
    def static_index(index: str, key_field: str = None, doc_type: str = None) \
            -> 'ElasticsearchEmitter':
        """
        Creates an emitter with static index which is invoked on every record to convert it to
        Elasticsearch actions.
        """
        JMapElasticsearchEmitter = get_gateway().jvm \
            .org.apache.flink.connector.elasticsearch.sink.MapElasticsearchEmitter
        j_emitter = JMapElasticsearchEmitter(index, doc_type, key_field, False)
        return ElasticsearchEmitter(j_emitter)

    @staticmethod
    def dynamic_index(index_field: str, key_field: str = None, doc_type: str = None) \
            -> 'ElasticsearchEmitter':
        """
        Creates an emitter with dynamic index which is invoked on every record to convert it to
        Elasticsearch actions.
        """
        JMapElasticsearchEmitter = get_gateway().jvm \
            .org.apache.flink.connector.elasticsearch.sink.MapElasticsearchEmitter
        j_emitter = JMapElasticsearchEmitter(index_field, doc_type, key_field, True)
        return ElasticsearchEmitter(j_emitter)


class ElasticsearchSinkBuilderBase(abc.ABC):
    """
    Base builder to construct a ElasticsearchSink.
    """

    @abc.abstractmethod
    def __init__(self):
        self._j_elasticsearch_sink_builder = None

    @abc.abstractmethod
    def get_http_host_class(self):
        """
        Gets the org.apache.http.HttpHost class which path is different in different Elasticsearch
        version.
        """
        pass

    def set_emitter(self, emitter: ElasticsearchEmitter) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the emitter which is invoked on every record to convert it to Elasticsearch actions.

        :param emitter: The emitter to process records into Elasticsearch actions.
        """
        self._j_elasticsearch_sink_builder.setEmitter(emitter._j_emitter)
        return self

    def set_hosts(self, hosts: Union[str, List[str]]) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the hosts where the Elasticsearch cluster nodes are reachable.
        """
        if not isinstance(hosts, list):
            hosts = [hosts]
        JHttpHost = self.get_http_host_class()
        j_http_hosts_list = [JHttpHost.create(x) for x in hosts]
        j_http_hosts_array = to_jarray(JHttpHost, j_http_hosts_list)
        self._j_elasticsearch_sink_builder.setHosts(j_http_hosts_array)
        return self

    def set_delivery_guarantee(self, delivery_guarantee: DeliveryGuarantee) \
            -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the wanted DeliveryGuarantee. The default delivery guarantee is DeliveryGuarantee#NONE
        """
        j_delivery_guarantee = delivery_guarantee._to_j_delivery_guarantee()
        self._j_elasticsearch_sink_builder.setDeliveryGuarantee(j_delivery_guarantee)
        return self

    def set_bulk_flush_max_actions(self, num_max_actions: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the maximum number of actions to buffer for each bulk request. You can pass -1 to
        disable it. The default flush size 1000.
        """
        self._j_elasticsearch_sink_builder.setBulkFlushMaxActions(num_max_actions)
        return self

    def set_bulk_flush_max_size_mb(self, max_size_mb: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the maximum size of buffered actions, in mb, per bulk request. You can pass -1 to
        disable it.
        """
        self._j_elasticsearch_sink_builder.setBulkFlushMaxSizeMb(max_size_mb)
        return self

    def set_bulk_flush_interval(self, interval_millis: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the bulk flush interval, in milliseconds. You can pass -1 to disable it.
        """
        self._j_elasticsearch_sink_builder.setBulkFlushInterval(interval_millis)
        return self

    def set_bulk_flush_backoff_strategy(self,
                                        flush_backoff_type: FlushBackoffType,
                                        max_retries: int,
                                        delay_millis: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the type of back off to use when flushing bulk requests. The default bulk flush back
        off type is FlushBackoffType#NONE.

        Sets the amount of delay between each backoff attempt when flushing bulk requests, in
        milliseconds.

        Sets the maximum number of retries for a backoff attempt when flushing bulk requests.
        """
        self._j_elasticsearch_sink_builder.setBulkFlushBackoffStrategy(
            flush_backoff_type._to_j_flush_backoff_type(), max_retries, delay_millis)
        return self

    def set_connection_username(self, username: str) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the username used to authenticate the connection with the Elasticsearch cluster.
        """
        self._j_elasticsearch_sink_builder.setConnectionUsername(username)
        return self

    def set_connection_password(self, password: str) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the password used to authenticate the connection with the Elasticsearch cluster.
        """
        self._j_elasticsearch_sink_builder.setConnectionPassword(password)
        return self

    def set_connection_path_prefix(self, prefix: str) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets a prefix which used for every REST communication to the Elasticsearch cluster.
        """
        self._j_elasticsearch_sink_builder.setConnectionPathPrefix(prefix)
        return self

    def set_connection_request_timeout(self, timeout: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the timeout for requesting the connection of the Elasticsearch cluster from the
        connection manager.
        """
        self._j_elasticsearch_sink_builder.setConnectionRequestTimeout(timeout)
        return self

    def set_connection_timeout(self, timeout: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the timeout for establishing a connection of the Elasticsearch cluster.
        """
        self._j_elasticsearch_sink_builder.setConnectionTimeout(timeout)
        return self

    def set_socket_timeout(self, timeout: int) -> 'ElasticsearchSinkBuilderBase':
        """
        Sets the timeout for waiting for data or, put differently, a maximum period inactivity
        between two consecutive data packets.
        """
        self._j_elasticsearch_sink_builder.setSocketTimeout(timeout)
        return self

    def build(self) -> 'ElasticsearchSink':
        """
        Constructs the ElasticsearchSink with the properties configured this builder.
        """
        return ElasticsearchSink(self._j_elasticsearch_sink_builder.build())


class Elasticsearch6SinkBuilder(ElasticsearchSinkBuilderBase):
    """
    Builder to construct an Elasticsearch 6 compatible ElasticsearchSink.

    The following example shows the minimal setup to create a ElasticsearchSink that submits
    actions on checkpoint or the default number of actions was buffered (1000).

    Example:
    ::

        >>> sink = Elasticsearch6SinkBuilder() \\
        ...     .set_hosts('localhost:9200') \\
        ...     .set_emitter(ElasticsearchEmitter.static_index("user", "key_col")) \\
        ...     .build()
    """

    def __init__(self):
        self._j_elasticsearch_sink_builder = get_gateway().jvm \
            .org.apache.flink.connector.elasticsearch.sink.Elasticsearch6SinkBuilder()

    def get_http_host_class(self):
        return get_gateway().jvm.org.apache.flink.elasticsearch6.shaded.org.apache.http.HttpHost


class Elasticsearch7SinkBuilder(ElasticsearchSinkBuilderBase):
    """
    Builder to construct an Elasticsearch 7 compatible ElasticsearchSink.

    The following example shows the minimal setup to create a ElasticsearchSink that submits
    actions on checkpoint or the default number of actions was buffered (1000).

    Example:
    ::

        >>> sink = Elasticsearch7SinkBuilder() \\
        ...     .set_hosts('localhost:9200') \\
        ...     .set_emitter(ElasticsearchEmitter.dynamic_index("index_col", "key_col")) \\
        ...     .build()
    """

    def __init__(self):
        self._j_elasticsearch_sink_builder = get_gateway().jvm \
            .org.apache.flink.connector.elasticsearch.sink.Elasticsearch7SinkBuilder()

    def get_http_host_class(self):
        return get_gateway().jvm.org.apache.flink.elasticsearch7.shaded.org.apache.http.HttpHost


class ElasticsearchSink(Sink):
    """
    Flink Sink to insert or update data in an Elasticsearch index. The sink supports the following
    delivery guarantees.

    DeliveryGuarantee.NONE does not provide any guarantees: actions are flushed to Elasticsearch
    only depending on the configurations of the bulk processor. In case of a failure, it might
    happen that actions are lost if the bulk processor still has buffered actions.

    DeliveryGuarantee.AT_LEAST_ONCE on a checkpoint the sink will wait until all buffered actions
    are flushed to and acknowledged by Elasticsearch. No actions will be lost but actions might be
    sent to Elasticsearch multiple times when Flink restarts. These additional requests may cause
    inconsistent data in ElasticSearch right after the restart, but eventually everything will be
    consistent again.
    """
    def __init__(self, j_elasticsearch_sink):
        super(ElasticsearchSink, self).__init__(sink=j_elasticsearch_sink)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from enum import Enum

from pyflink.java_gateway import get_gateway

__all__ = ['InputDependencyConstraint']


class InputDependencyConstraint(Enum):
    """
    This constraint indicates when a task should be scheduled considering its inputs status.

    :data:`ANY`:

    Schedule the task if any input is consumable.

    :data:`ALL`:

    Schedule the task if all the inputs are consumable.
    """

    ANY = 0
    ALL = 1

    @staticmethod
    def _from_j_input_dependency_constraint(j_input_dependency_constraint) \
            -> 'InputDependencyConstraint':
        return InputDependencyConstraint[j_input_dependency_constraint.name()]

    def _to_j_input_dependency_constraint(self):
        gateway = get_gateway()
        JInputDependencyConstraint = gateway.jvm.org.apache.flink.api.common \
            .InputDependencyConstraint
        return getattr(JInputDependencyConstraint, self.name)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from abc import abstractmethod, ABC
from typing import Generic, TypeVar

K = TypeVar('K')
N = TypeVar('N')


class InternalTimerService(Generic[N], ABC):
    """
    Interface for working with time and timers.

    This is the internal version of TimerService that allows to specify a key and a namespace to
    which timers should be scoped.
    """

    @abstractmethod
    def current_processing_time(self):
        """
        Returns the current processing time.
        """
        pass

    @abstractmethod
    def current_watermark(self):
        """
        Returns the current event-time watermark.
        """
        pass

    @abstractmethod
    def register_processing_time_timer(self, namespace: N, t: int):
        """
        Registers a timer to be fired when processing time passes the given time. The namespace you
        pass here will be provided when the timer fires.

        :param namespace: The namespace you pass here will be provided when the timer fires.
        :param t: The processing time of the timer to be registered.
        """
        pass

    @abstractmethod
    def register_event_time_timer(self, namespace: N, t: int):
        """
        Registers a timer to be fired when event time watermark passes the given time. The namespace
        you pass here will be provided when the timer fires.

        :param namespace: The namespace you pass here will be provided when the timer fires.
        :param t: The event time of the timer to be registered.
        """
        pass

    def delete_processing_time_timer(self, namespace: N, t: int):
        """
        Deletes the timer for the given key and namespace.

        :param namespace: The namespace you pass here will be provided when the timer fires.
        :param t: The given trigger time of timer to be deleted.
        """
        pass

    def delete_event_time_timer(self, namespace: N, t: int):
        """
        Deletes the timer for the given key and namespace.

        :param namespace: The namespace you pass here will be provided when the timer fires.
        :param t: The given trigger time of timer to be deleted.
        """
        pass


class InternalTimer(Generic[K, N], ABC):

    @abstractmethod
    def get_timestamp(self) -> int:
        """
        Returns the timestamp of the timer. This value determines the point in time when the timer
        will fire.
        """
        pass

    @abstractmethod
    def get_key(self) -> K:
        """
        Returns the key that is bound to this timer.
        """
        pass

    @abstractmethod
    def get_namespace(self) -> N:
        """
        Returns the namespace that is bound to this timer.
        :return:
        """
        pass



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.datastream.connectors import Source
from pyflink.java_gateway import get_gateway

__all__ = [
    'NumberSequenceSource'
]


class NumberSequenceSource(Source):
    """
    A data source that produces a sequence of numbers (longs). This source is useful for testing and
    for cases that just need a stream of N events of any kind.

    The source splits the sequence into as many parallel sub-sequences as there are parallel
    source readers. Each sub-sequence will be produced in order. Consequently, if the parallelism is
    limited to one, this will produce one sequence in order.

    This source is always bounded. For very long sequences (for example over the entire domain of
    long integer values), user may want to consider executing the application in a streaming manner,
    because, despite the fact that the produced stream is bounded, the end bound is pretty far away.
    """

    def __init__(self, start: int, end: int):
        """
        Creates a new NumberSequenceSource that produces parallel sequences covering the
        range start to end (both boundaries are inclusive).
        """
        JNumberSequenceSource = get_gateway().jvm.org.apache.flink.api.connector.source.lib.\
            NumberSequenceSource
        j_seq_source = JNumberSequenceSource(start, end)
        super(NumberSequenceSource, self).__init__(source=j_seq_source)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Union

from pyflink.datastream.state import ListState, MapState
from pyflink.fn_execution.coders import from_proto, PickleCoder
from pyflink.fn_execution.internal_state import InternalListState, InternalMapState
from pyflink.fn_execution.utils.operation_utils import is_built_in_function, load_aggregate_function
from pyflink.table import FunctionContext
from pyflink.table.data_view import ListView, MapView, DataView


def extract_data_view_specs_from_accumulator(current_index, accumulator):
    # for built in functions we extract the data view specs from their accumulator
    i = -1
    extracted_specs = []
    for field in accumulator:
        i += 1
        # TODO: infer the coder from the input types and output type of the built-in functions
        if isinstance(field, MapView):
            extracted_specs.append(MapViewSpec(
                "builtInAgg%df%d" % (current_index, i), i, PickleCoder(), PickleCoder()))
        elif isinstance(field, ListView):
            extracted_specs.append(ListViewSpec(
                "builtInAgg%df%d" % (current_index, i), i, PickleCoder()))
    return extracted_specs


def extract_data_view_specs(udfs):
    extracted_udf_data_view_specs = []
    current_index = -1
    for udf in udfs:
        current_index += 1
        udf_data_view_specs_proto = udf.specs
        if not udf_data_view_specs_proto:
            if is_built_in_function(udf.payload):
                built_in_function = load_aggregate_function(udf.payload)
                accumulator = built_in_function.create_accumulator()
                extracted_udf_data_view_specs.append(
                    extract_data_view_specs_from_accumulator(current_index, accumulator))
            else:
                extracted_udf_data_view_specs.append([])
        else:
            extracted_specs = []
            for spec_proto in udf_data_view_specs_proto:
                state_id = spec_proto.name
                field_index = spec_proto.field_index
                if spec_proto.HasField("list_view"):
                    element_coder = from_proto(spec_proto.list_view.element_type)
                    extracted_specs.append(ListViewSpec(state_id, field_index, element_coder))
                elif spec_proto.HasField("map_view"):
                    key_coder = from_proto(spec_proto.map_view.key_type)
                    value_coder = from_proto(spec_proto.map_view.value_type)
                    extracted_specs.append(
                        MapViewSpec(state_id, field_index, key_coder, value_coder))
                else:
                    raise Exception("Unsupported data view spec type: " + spec_proto.type)
            extracted_udf_data_view_specs.append(extracted_specs)
    if all([len(i) == 0 for i in extracted_udf_data_view_specs]):
        return []
    return extracted_udf_data_view_specs


N = TypeVar('N')


class StateDataView(DataView, Generic[N]):
    @abstractmethod
    def set_current_namespace(self, namespace: N):
        """
        Sets current namespace for state.
        """
        pass


class StateListView(ListView, StateDataView[N], ABC):

    def __init__(self, list_state: Union[ListState, InternalListState]):
        super().__init__()
        self._list_state = list_state

    def get(self):
        return self._list_state.get()

    def add(self, value):
        self._list_state.add(value)

    def add_all(self, values):
        self._list_state.add_all(values)

    def clear(self):
        self._list_state.clear()

    def __hash__(self) -> int:
        return hash([i for i in self.get()])


class KeyedStateListView(StateListView[N]):
    """
    KeyedStateListView is a default implementation of StateListView whose underlying
    representation is a keyed state.
    """

    def __init__(self, list_state: ListState):
        super(KeyedStateListView, self).__init__(list_state)

    def set_current_namespace(self, namespace: N):
        raise Exception("KeyedStateListView doesn't support set_current_namespace")


class NamespacedStateListView(StateListView[N]):
    """
    NamespacedStateListView is a StateListView whose underlying representation is a keyed and
    namespaced state. It also supports changing current namespace.
    """

    def __init__(self, list_state: InternalListState):
        super(NamespacedStateListView, self).__init__(list_state)

    def set_current_namespace(self, namespace: N):
        self._list_state.set_current_namespace(namespace)


class StateMapView(MapView, StateDataView[N], ABC):

    def __init__(self, map_state: Union[MapState, InternalMapState]):
        super().__init__()
        self._map_state = map_state

    def get(self, key):
        return self._map_state.get(key)

    def put(self, key, value) -> None:
        self._map_state.put(key, value)

    def put_all(self, dict_value) -> None:
        self._map_state.put_all(dict_value)

    def remove(self, key) -> None:
        self._map_state.remove(key)

    def contains(self, key) -> bool:
        return self._map_state.contains(key)

    def items(self):
        return self._map_state.items()

    def keys(self):
        return self._map_state.keys()

    def values(self):
        return self._map_state.values()

    def is_empty(self) -> bool:
        return self._map_state.is_empty()

    def clear(self) -> None:
        return self._map_state.clear()


class KeyedStateMapView(StateMapView[N]):
    """
    KeyedStateMapView is a default implementation of StateMapView whose underlying
    representation is a keyed state.
    """

    def __init__(self, map_state: MapState):
        super(KeyedStateMapView, self).__init__(map_state)

    def set_current_namespace(self, namespace: N):
        raise Exception("KeyedStateMapView doesn't support set_current_namespace")


class NamespacedStateMapView(StateMapView[N]):
    """
    NamespacedStateMapView is a StateMapView whose underlying representation is a keyed and
    namespaced state. It also supports changing current namespace.
    """

    def __init__(self, map_state: InternalMapState):
        super(NamespacedStateMapView, self).__init__(map_state)

    def set_current_namespace(self, namespace: N):
        self._map_state.set_current_namespace(namespace)


class DataViewSpec(object):

    def __init__(self, state_id, field_index):
        self.state_id = state_id
        self.field_index = field_index


class ListViewSpec(DataViewSpec):

    def __init__(self, state_id, field_index, element_coder):
        super(ListViewSpec, self).__init__(state_id, field_index)
        self.element_coder = element_coder


class MapViewSpec(DataViewSpec):

    def __init__(self, state_id, field_index, key_coder, value_coder):
        super(MapViewSpec, self).__init__(state_id, field_index)
        self.key_coder = key_coder
        self.value_coder = value_coder


class StateDataViewStore(ABC):
    """
    This interface contains methods for registering StateDataView with a managed store.
    """

    def __init__(self,
                 function_context: FunctionContext,
                 keyed_state_backend):
        self._function_context = function_context
        self._keyed_state_backend = keyed_state_backend

    def get_runtime_context(self):
        return self._function_context

    @abstractmethod
    def get_state_list_view(self, state_name, element_coder):
        """
        Creates a state list view.

        :param state_name: The name of underlying state of the list view.
        :param element_coder: The element coder
        :return: a keyed list state
        """
        pass

    @abstractmethod
    def get_state_map_view(self, state_name, key_coder, value_coder):
        """
        Creates a state map view.

        :param state_name: The name of underlying state of the map view.
        :param key_coder: The key coder
        :param value_coder: The value coder
        :return: a keyed map state
        """
        pass


class PerKeyStateDataViewStore(StateDataViewStore):
    """
    Default implementation of StateDataViewStore.
    """

    def __init__(self,
                 function_context: FunctionContext,
                 keyed_state_backend):
        super(PerKeyStateDataViewStore, self).__init__(function_context, keyed_state_backend)

    def get_state_list_view(self, state_name, element_coder):
        return KeyedStateListView(
            self._keyed_state_backend.get_list_state(state_name, element_coder))

    def get_state_map_view(self, state_name, key_coder, value_coder):
        return KeyedStateMapView(
            self._keyed_state_backend.get_map_state(state_name, key_coder, value_coder))


class PerWindowStateDataViewStore(StateDataViewStore):
    """
    An implementation of StateDataViewStore for window aggregates which forwards the state
    registration to an underlying RemoteKeyedStateBackend. The state created by this store has the
    ability to switch window namespaces.
    """

    def __init__(self,
                 function_context: FunctionContext,
                 keyed_state_backend):
        super(PerWindowStateDataViewStore, self).__init__(function_context, keyed_state_backend)

    def get_state_list_view(self, state_name, element_coder):
        return NamespacedStateListView(
            self._keyed_state_backend.get_list_state(state_name, element_coder))

    def get_state_map_view(self, state_name, key_coder, value_coder):
        return NamespacedStateMapView(
            self._keyed_state_backend.get_map_state(state_name, key_coder, value_coder))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from py4j.java_gateway import get_java_class, JavaObject, java_import

from pyflink.common.io import InputFormat
from pyflink.common.serialization import BulkWriterFactory, SerializationSchema, \
    DeserializationSchema
from pyflink.common.typeinfo import TypeInformation

from pyflink.datastream.utils import ResultTypeQueryable
from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import get_field_value, load_java_class

__all__ = [
    'AvroSchema',
    'GenericRecordAvroTypeInfo',
    'AvroInputFormat',
    'AvroBulkWriters',
    'AvroRowDeserializationSchema',
    'AvroRowSerializationSchema'
]


class AvroSchema(object):
    """
    Avro Schema class contains Java org.apache.avro.Schema.

    .. versionadded:: 1.16.0
    """

    def __init__(self, j_schema):
        self._j_schema = j_schema
        self._schema_string = None

    def __str__(self):
        if self._schema_string is None:
            self._schema_string = get_field_value(self._j_schema, 'schema').toString()
        return self._schema_string

    @staticmethod
    def parse_string(json_schema: str) -> 'AvroSchema':
        """
        Parse JSON string as Avro Schema.

        :param json_schema: JSON represented schema string.
        :return: the Avro Schema.
        """
        JSchema = get_gateway().jvm.org.apache.flink.avro.shaded.org.apache.avro.Schema
        return AvroSchema(JSchema.Parser().parse(json_schema))

    @staticmethod
    def parse_file(file_path: str) -> 'AvroSchema':
        """
        Parse a schema definition file as Avro Schema.

        :param file_path: path to schema definition file.
        :return: the Avro Schema.
        """
        jvm = get_gateway().jvm
        j_file = jvm.java.io.File(file_path)
        JSchema = jvm.org.apache.flink.avro.shaded.org.apache.avro.Schema
        return AvroSchema(JSchema.Parser().parse(j_file))


class GenericRecordAvroTypeInfo(TypeInformation):
    """
    A :class:`TypeInformation` of Avro's GenericRecord, including the schema. This is a wrapper of
    Java org.apache.flink.formats.avro.typeutils.GenericRecordAvroTypeInfo.

    Note that this type cannot be used as the type_info of data in
    :meth:`StreamExecutionEnvironment.from_collection`.

    .. versionadded::1.16.0
    """

    def __init__(self, schema: 'AvroSchema'):
        super(GenericRecordAvroTypeInfo, self).__init__()
        self._schema = schema
        self._j_typeinfo = get_gateway().jvm.org.apache.flink.formats.avro.typeutils \
            .GenericRecordAvroTypeInfo(schema._j_schema)

    def get_java_type_info(self) -> JavaObject:
        return self._j_typeinfo


class AvroInputFormat(InputFormat, ResultTypeQueryable):
    """
    Provides a FileInputFormat for Avro records.

    Example:
    ::

        >>> env = StreamExecutionEnvironment.get_execution_environment()
        >>> schema = AvroSchema.parse_string(JSON_SCHEMA)
        >>> ds = env.create_input(AvroInputFormat(FILE_PATH, schema))

    .. versionadded:: 1.16.0
    """

    def __init__(self, path: str, schema: 'AvroSchema'):
        """
        :param path: The path to Avro data file.
        :param schema: The :class:`AvroSchema` of generic record.
        """
        jvm = get_gateway().jvm
        j_avro_input_format = jvm.org.apache.flink.formats.avro.AvroInputFormat(
            jvm.org.apache.flink.core.fs.Path(path),
            get_java_class(jvm.org.apache.flink.avro.shaded.org.apache.avro.generic.GenericRecord)
        )
        super(AvroInputFormat, self).__init__(j_avro_input_format)
        self._type_info = GenericRecordAvroTypeInfo(schema)

    def get_produced_type(self) -> GenericRecordAvroTypeInfo:
        return self._type_info


class AvroBulkWriters(object):
    """
    Convenience builder to create :class:`~pyflink.common.serialization.BulkWriterFactory` for
    Avro types.

    .. versionadded:: 1.16.0
    """

    @staticmethod
    def for_generic_record(schema: 'AvroSchema') -> 'BulkWriterFactory':
        """
        Creates an AvroWriterFactory that accepts and writes Avro generic types. The Avro writers
        will use the given schema to build and write the records.

        Note that to make this works in PyFlink, you need to declare the output type of the
        predecessor before FileSink to be :class:`~GenericRecordAvroTypeInfo`, and the predecessor
        cannot be :meth:`StreamExecutionEnvironment.from_collection`, you can add a pass-through map
        function before the sink, as the example shown below.

        The Python data records should match the Avro schema, and have the same behavior with
        vanilla Python data structure, e.g. an object for Avro array should behave like Python list,
        an object for Avro map should behave like Python dict.

        Example:
        ::

            >>> env = StreamExecutionEnvironment.get_execution_environment()
            >>> schema = AvroSchema(JSON_SCHEMA)
            >>> avro_type_info = GenericRecordAvroTypeInfo(schema)
            >>> ds = env.from_collection([{'array': [1, 2]}], type_info=Types.PICKLED_BYTE_ARRAY())
            >>> sink = FileSink.for_bulk_format(
            ...     OUTPUT_DIR, AvroBulkWriters.for_generic_record(schema)).build()
            >>> # A map to indicate its Avro type info is necessary for serialization
            >>> ds.map(lambda e: e, output_type=GenericRecordAvroTypeInfo(schema)) \\
            ...     .sink_to(sink)

        :param schema: The avro schema.
        :return: The BulkWriterFactory to write generic records into avro files.
        """
        jvm = get_gateway().jvm
        j_bulk_writer_factory = jvm.org.apache.flink.formats.avro.AvroWriters.forGenericRecord(
            schema._j_schema
        )
        return BulkWriterFactory(j_bulk_writer_factory)


class AvroRowDeserializationSchema(DeserializationSchema):
    """
    Deserialization schema from Avro bytes to Row. Deserializes the byte[] messages into (nested)
    Flink rows. It converts Avro types into types that are compatible with Flink's Table & SQL API.

    Projects with Avro records containing logical date/time types need to add a JodaTime dependency.
    """
    def __init__(self, record_class: str = None, avro_schema_string: str = None):
        """
        Creates an Avro deserialization schema for the given specific record class or Avro schema
        string. Having the concrete Avro record class might improve performance.

        :param record_class: Avro record class used to deserialize Avro's record to Flink's row.
        :param avro_schema_string: Avro schema string to deserialize Avro's record to Flink's row.
        """

        if avro_schema_string is None and record_class is None:
            raise TypeError("record_class or avro_schema_string should be specified.")
        j_deserialization_schema = None
        if record_class is not None:
            gateway = get_gateway()
            java_import(gateway.jvm, record_class)
            j_record_class = load_java_class(record_class)
            JAvroRowDeserializationSchema = get_gateway().jvm \
                .org.apache.flink.formats.avro.AvroRowDeserializationSchema
            j_deserialization_schema = JAvroRowDeserializationSchema(j_record_class)

        elif avro_schema_string is not None:
            JAvroRowDeserializationSchema = get_gateway().jvm \
                .org.apache.flink.formats.avro.AvroRowDeserializationSchema
            j_deserialization_schema = JAvroRowDeserializationSchema(avro_schema_string)

        super(AvroRowDeserializationSchema, self).__init__(j_deserialization_schema)


class AvroRowSerializationSchema(SerializationSchema):
    """
    Serialization schema that serializes to Avro binary format.
    """

    def __init__(self, record_class: str = None, avro_schema_string: str = None):
        """
        Creates AvroSerializationSchema that serializes SpecificRecord using provided schema or
        record class.

        :param record_class: Avro record class used to serialize  Flink's row to Avro's record.
        :param avro_schema_string: Avro schema string to serialize Flink's row to Avro's record.
        """
        if avro_schema_string is None and record_class is None:
            raise TypeError("record_class or avro_schema_string should be specified.")

        j_serialization_schema = None
        if record_class is not None:
            gateway = get_gateway()
            java_import(gateway.jvm, record_class)
            j_record_class = load_java_class(record_class)
            JAvroRowSerializationSchema = get_gateway().jvm \
                .org.apache.flink.formats.avro.AvroRowSerializationSchema
            j_serialization_schema = JAvroRowSerializationSchema(j_record_class)

        elif avro_schema_string is not None:
            JAvroRowSerializationSchema = get_gateway().jvm \
                .org.apache.flink.formats.avro.AvroRowSerializationSchema
            j_serialization_schema = JAvroRowSerializationSchema(avro_schema_string)

        super(AvroRowSerializationSchema, self).__init__(j_serialization_schema)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""
Entry point classes of Flink DataStream API:

    - :class:`StreamExecutionEnvironment`:
      The context in which a streaming program is executed.
    - :class:`DataStream`:
      Represents a stream of elements of the same type. A DataStream can be transformed
      into another DataStream by applying a transformation.
    - :class:`KeyedStream`:
      Represents a :class:`DataStream` where elements are partitioned by key using a
      provided KeySelector.
    - :class:`WindowedStream`:
      Represents a data stream where elements are grouped by key, and for each
      key, the stream of elements is split into windows based on a WindowAssigner. Window emission
      is triggered based on a Trigger.
    - :class:`ConnectedStreams`:
      Represent two connected streams of (possibly) different data types. Connected
      streams are useful for cases where operations on one stream directly affect the operations on
      the other stream, usually via shared state between the streams.
    - :class:`BroadcastStream`:
      Represent a stream with :class:`state.BroadcastState` (s).
    - :class:`BroadcastConnectedStream`:
      Represents the result of connecting a keyed or non-keyed stream, with a
      :class:`BroadcastStream` with :class:`state.BroadcastState` (s)

Functions used to transform a :class:`DataStream` into another :class:`DataStream`:

    - :class:`MapFunction`:
      Performs a map transformation of a :class:`DataStream` at element wise.
    - :class:`CoMapFunction`:
      Performs a map transformation over two connected streams.
    - :class:`FlatMapFunction`:
      Performs a flatmap transformation of a :class:`DataStream` which produces zero, one, or more
      elements for each input element.
    - :class:`CoFlatMapFunction`:
      Performs a flatmap transformation over two connected streams.
    - :class:`FilterFunction`:
      A filter function is a predicate applied individually to each record.
    - :class:`ReduceFunction`:
      Combines groups of elements to a single value.
    - :class:`ProcessFunction`:
      Similar to :class:`FlatMapFunction`, except that it could access the current timestamp and
      watermark in :class:`ProcessFunction`.
    - :class:`KeyedProcessFunction`:
      Similar to :class:`ProcessFunction`, except that it was applied to a :class:`KeyedStream` and
      could register event-time and processing-time timers.
    - :class:`CoProcessFunction`:
      Similar to :class:`CoFlatMapFunction`, except that it could access the current timestamp and
      watermark in :class:`CoProcessFunction`.
    - :class:`KeyedCoProcessFunction`:
      Similar to :class:`CoProcessFunction`, except that it was applied to a keyed
      :class:`ConnectedStreams` and could register event-time and processing-time timers.
    - :class:`WindowFunction`:
      Base interface for functions that are evaluated over keyed (grouped) windows.
    - :class:`ProcessWindowFunction`:
      Similar to :class:`WindowFunction`, except that it could access a context for retrieving extra
      information such as the current timestamp, the watermark, etc.
    - :class:`AggregateFunction`:
      Base class for a user-defined aggregate function.
    - :class:`BroadcastProcessFunction`:
      A function to be applied to a :class:`BroadcastConnectedStream` that connects
      :class:`BroadcastStream`, i.e. a stream with broadcast state, with a non-keyed
      :class:`DataStream`.
    - :class:`KeyedBroadcastProcessFunction`:
      A function to be applied to a :class:`BroadcastConnectedStream` that connects
      :class:`BroadcastStream`, i.e. a stream with broadcast state, with a :class:`KeyedStream`.
    - :class:`RuntimeContext`:
      Contains information about the context in which functions are executed. Each
      parallel instance of the function will have a context through which it can access static
      contextual information (such as the current parallelism), etc.

Classes to define window:

    - :class:`Window`:
      A grouping of elements into finite buckets.
    - :class:`TimeWindow`:
      A grouping of elements according to a time interval from start (inclusive) to end (exclusive).
    - :class:`CountWindow`:
      A grouping of elements according to element count from start (inclusive) to end (exclusive).
    - :class:`GlobalWindow`:
      The window into which all data is placed.
    - :class:`WindowAssigner`:
      Assigns zero or more :class:`Window` to an element.
    - :class:`MergingWindowAssigner`:
      A :class:`WindowAssigner` that can merge windows.
    - :class:`TriggerResult`:
      Result type for trigger methods. This determines what happens with the window, for example
      whether the window function should be called, or the window should be discarded.
    - :class:`Trigger`:
      Determines when a pane of a window should be evaluated to emit the results for that
      part of the window.

Classes to define the behavior of checkpoint and state backend:

    - :class:`CheckpointingMode`:
      Defines what consistency guarantees the system gives in the presence of failures.
    - :class:`CheckpointConfig`:
      Configuration that captures all checkpointing related settings.
    - :class:`StateBackend`:
      Base class of the state backends which define how the state of a streaming application is
      stored locally within the cluster. Different state backends store their state in different
      fashions, and use different data structures to hold the state of a running application.
    - :class:`HashMapStateBackend`:
      Holds the working state in the memory (JVM heap) of the TaskManagers and
      checkpoints based on the configured :class:`CheckpointStorage`.
    - :class:`EmbeddedRocksDBStateBackend`:
      Stores its state in an embedded `RocksDB` instance. This state backend can store very large
      state that exceeds memory and spills to local disk.
    - :class:`CustomStateBackend`:
      A wrapper of customized java state backend.
    - :class:`JobManagerCheckpointStorage`:
      Checkpoints state directly to the JobManager's memory (hence the name), but savepoints will
      be persisted to a file system.
    - :class:`FileSystemCheckpointStorage`:
      Checkpoints state as files to a file system. Each checkpoint individually will store all its
      files in a subdirectory that includes the checkpoint number, such as
      `hdfs://namenode:port/flink-checkpoints/chk-17/`.
    - :class:`CustomCheckpointStorage`:
      A wrapper of customized java checkpoint storage.

Classes for state operations:

    - :class:`state.ValueState`:
      Interface for partitioned single-value state. The value can be retrieved or updated.
    - :class:`state.ListState`:
      Interface for partitioned list state in Operations. The state is accessed and modified by
      user functions, and checkpointed consistently by the system as part of the distributed
      snapshots.
    - :class:`state.MapState`:
      Interface for partitioned key-value state. The key-value pair can be added, updated and
      retrieved.
    - :class:`state.ReducingState`:
      Interface for reducing state. Elements can be added to the state, they will be combined using
      a :class:`ReduceFunction`. The current state can be inspected.
    - :class:`state.AggregatingState`:
      Interface for aggregating state, based on an :class:`AggregateFunction`. Elements that are
      added to this type of state will be eagerly pre-aggregated using a given AggregateFunction.
    - :class:`state.BroadcastState`:
      A type of state that can be created to store the state of a :class:`BroadcastStream`. This
      state assumes that the same elements are sent to all instances of an operator.
    - :class:`state.ReadOnlyBroadcastState`:
      A read-only view of the :class:`state.BroadcastState`.
    - :class:`state.StateTtlConfig`:
      Configuration of state TTL logic.

Classes to define source & sink:

    - :class:`connectors.elasticsearch.ElasticsearchSink`:
      A sink for publishing data into Elasticsearch 6 or Elasticsearch 7.
    - :class:`connectors.kafka.FlinkKafkaConsumer`:
      A streaming data source that pulls a parallel data stream from Apache Kafka.
    - :class:`connectors.kafka.FlinkKafkaProducer`:
      A streaming data sink to produce data into a Kafka topic.
    - :class:`connectors.kafka.KafkaSource`:
      The new API to read data in parallel from Apache Kafka.
    - :class:`connectors.kafka.KafkaSink`:
      The new API to write data into to Apache Kafka topics.
    - :class:`connectors.file_system.FileSource`:
      A unified data source that reads files - both in batch and in streaming mode.
      This source supports all (distributed) file systems and object stores that can be accessed via
      the Flink's FileSystem class.
    - :class:`connectors.file_system.FileSink`:
      A unified sink that emits its input elements to FileSystem files within buckets. This
      sink achieves exactly-once semantics for both BATCH and STREAMING.
    - :class:`connectors.file_system.StreamingFileSink`:
      Sink that emits its input elements to files within buckets. This is integrated with the
      checkpointing mechanism to provide exactly once semantics.
    - :class:`connectors.number_seq.NumberSequenceSource`:
      A data source that produces a sequence of numbers (longs). This source is useful for testing
      and for cases that just need a stream of N events of any kind.
    - :class:`connectors.jdbc.JdbcSink`:
      A data sink to produce data into an external storage using JDBC.
    - :class:`connectors.pulsar.PulsarSource`:
      A streaming data source that pulls a parallel data stream from Pulsar.
    - :class:`connectors.pulsar.PulsarSink`:
      A streaming data sink to produce data into Pulsar.
    - :class:`connectors.rabbitmq.RMQSource`:
      A streaming data source that pulls a parallel data stream from RabbitMQ.
    - :class:`connectors.rabbitmq.RMQSink`:
      A Sink for publishing data into RabbitMQ.
    - :class:`connectors.cassandra.CassandraSink`:
      A Sink for publishing data into Cassandra.
    - :class:`connectors.kinesis.FlinkKinesisConsumer`:
      A streaming data source that pulls a parallel data stream from Kinesis.
    - :class:`connectors.kinesis.KinesisStreamsSink`:
      A Kinesis Data Streams (KDS) Sink that performs async requests against a destination stream
      using the buffering protocol.
    - :class:`connectors.kinesis.KinesisFirehoseSink`:
      A Kinesis Data Firehose (KDF) Sink that performs async requests against a destination delivery
      stream using the buffering protocol.
    - :class:`connectors.hybrid_source.HybridSource`:
      A Hybrid source that switches underlying sources based on configured source chain.


Classes to define formats used together with source & sink:

    - :class:`formats.csv.CsvReaderFormat`:
      A :class:`~connectors.file_system.StreamFormat` to read CSV files into Row data.
    - :class:`formats.csv.CsvBulkWriter`:
      Creates :class:`~pyflink.common.serialization.BulkWriterFactory` to write Row data into CSV
      files.
    - :class:`formats.avro.GenericRecordAvroTypeInfo`:
      A :class:`~pyflink.common.typeinfo.TypeInformation` to indicate vanilla Python records will be
      translated to GenericRecordAvroTypeInfo on the Java side.
    - :class:`formats.avro.AvroInputFormat`:
      An InputFormat to read avro files in a streaming fashion.
    - :class:`formats.avro.AvroWriters`:
      A class to provide :class:`~pyflink.common.serialization.BulkWriterFactory` to write vanilla
      Python objects into avro files in a batch fashion.
    - :class:`formats.parquet.ParquetColumnarRowInputFormat`:
      A :class:`~connectors.file_system.BulkFormat` to read columnar parquet files into Row data in
      a batch-processing fashion.
    - :class:`formats.parquet.ParquetBulkWriters`:
      Convenient builder to create a :class:`~pyflink.common.serialization.BulkWriterFactory` that
      writes Rows with a defined RowType into Parquet files in a batch fashion.
    - :class:`formats.parquet.AvroParquetReaders`:
      A convenience builder to create reader format that reads individual Avro records from a
      Parquet stream. Only GenericRecord is supported in PyFlink.
    - :class:`formats.parquet.AvroParquetWriters`:
      Convenience builder to create ParquetWriterFactory instances for Avro types. Only
      GenericRecord is supported in PyFlink.
    - :class:`formats.orc.OrcBulkWriters`:
      Convenient builder to create a :class:`~pyflink.common.serialization.BulkWriterFactory` that
      writes Row records with a defined :class:`RowType` into Orc files.

Other important classes:

    - :class:`TimeDomain`:
      Specifies whether a firing timer is based on event time or processing time.
    - :class:`KeySelector`:
      The extractor takes an object and returns the deterministic key for that object.
    - :class:`Partitioner`:
      Function to implement a custom partition assignment for keys.
    - :class:`SinkFunction`:
      Interface for implementing user defined sink functionality.
    - :class:`SourceFunction`:
      Interface for implementing user defined source functionality.
    - :class:`OutputTag`:
      Tag with a name and type for identifying side output of an operator
"""
from pyflink.datastream.checkpoint_config import CheckpointConfig
from pyflink.datastream.externalized_checkpoint_retention import ExternalizedCheckpointRetention
from pyflink.datastream.checkpointing_mode import CheckpointingMode
from pyflink.datastream.data_stream import DataStream, KeyedStream, WindowedStream, \
    ConnectedStreams, DataStreamSink, BroadcastStream, BroadcastConnectedStream
from pyflink.datastream.execution_mode import RuntimeExecutionMode
from pyflink.datastream.functions import (MapFunction, CoMapFunction, FlatMapFunction,
                                          CoFlatMapFunction, ReduceFunction, RuntimeContext,
                                          KeySelector, FilterFunction, Partitioner, SourceFunction,
                                          SinkFunction, CoProcessFunction, KeyedProcessFunction,
                                          KeyedCoProcessFunction, AggregateFunction, WindowFunction,
                                          ProcessWindowFunction, BroadcastProcessFunction,
                                          KeyedBroadcastProcessFunction)
from pyflink.datastream.slot_sharing_group import SlotSharingGroup, MemorySize
from pyflink.datastream.state_backend import (StateBackend, CustomStateBackend,
                                              PredefinedOptions, HashMapStateBackend,
                                              EmbeddedRocksDBStateBackend)
from pyflink.datastream.checkpoint_storage import (CheckpointStorage, JobManagerCheckpointStorage,
                                                   FileSystemCheckpointStorage,
                                                   CustomCheckpointStorage)
from pyflink.datastream.stream_execution_environment import StreamExecutionEnvironment
from pyflink.datastream.time_domain import TimeDomain
from pyflink.datastream.functions import ProcessFunction
from pyflink.datastream.timerservice import TimerService
from pyflink.datastream.window import Window, TimeWindow, CountWindow, WindowAssigner, \
    MergingWindowAssigner, TriggerResult, Trigger, GlobalWindow
from pyflink.datastream.output_tag import OutputTag

__all__ = [
    'StreamExecutionEnvironment',
    'DataStream',
    'KeyedStream',
    'WindowedStream',
    'ConnectedStreams',
    'BroadcastStream',
    'BroadcastConnectedStream',
    'DataStreamSink',
    'MapFunction',
    'CoMapFunction',
    'FlatMapFunction',
    'CoFlatMapFunction',
    'ReduceFunction',
    'FilterFunction',
    'ProcessFunction',
    'KeyedProcessFunction',
    'CoProcessFunction',
    'KeyedCoProcessFunction',
    'WindowFunction',
    'ProcessWindowFunction',
    'AggregateFunction',
    'BroadcastProcessFunction',
    'KeyedBroadcastProcessFunction',
    'RuntimeContext',
    'TimerService',
    'CheckpointingMode',
    'CheckpointConfig',
    'ExternalizedCheckpointRetention',
    'StateBackend',
    'HashMapStateBackend',
    'EmbeddedRocksDBStateBackend',
    'CustomStateBackend',
    'PredefinedOptions',
    'CheckpointStorage',
    'JobManagerCheckpointStorage',
    'FileSystemCheckpointStorage',
    'CustomCheckpointStorage',
    'RuntimeExecutionMode',
    'Window',
    'TimeWindow',
    'CountWindow',
    'GlobalWindow',
    'WindowAssigner',
    'MergingWindowAssigner',
    'TriggerResult',
    'Trigger',
    'TimeDomain',
    'KeySelector',
    'Partitioner',
    'SourceFunction',
    'SinkFunction',
    'SlotSharingGroup',
    'MemorySize',
    'OutputTag'
]



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import struct

from avro.errors import AvroTypeException, SchemaResolutionException
from avro.io import (
    BinaryDecoder,
    BinaryEncoder,
    DatumReader,
    DatumWriter,
    validate,
)

STRUCT_FLOAT = struct.Struct('>f')  # big-endian float
STRUCT_DOUBLE = struct.Struct('>d')  # big-endian double
STRUCT_INT = struct.Struct('>i')  # big-endian int
STRUCT_LONG_LONG = struct.Struct('>q')  # big-endian long long


class FlinkAvroBufferWrapper(object):

    def __init__(self):
        self._stream = None

    def switch_stream(self, stream):
        self._stream = stream

    def read(self, n=1):
        return self._stream.read(n)

    def write(self, data):
        return self._stream.write(bytes(data))


class FlinkAvroDecoder(BinaryDecoder):
    """
    Flink-customized basic type decoder, with some different encodings, e.g. Default avro encode int
    and long into variable-sized bytes, while Flink use fixed bytes in some places.
    """

    def __init__(self, reader):
        super().__init__(reader)

    def read_int(self):
        return STRUCT_INT.unpack(self.read(4))[0]

    def read_long(self):
        return STRUCT_LONG_LONG.unpack(self.read(8))[0]

    def read_var_long(self):
        """
        Flink implementation of variable-sized long serialization does not move sign to lower bit
        and flip the rest.
        """
        b = ord(self.read(1))
        n = b & 0x7F
        shift = 7
        while (b & 0x80) != 0:
            b = ord(self.read(1))
            n |= (b & 0x7F) << shift
            shift += 7
        return n

    def read_float(self):
        return STRUCT_FLOAT.unpack(self.read(4))[0]

    def read_double(self):
        return STRUCT_DOUBLE.unpack(self.read(8))[0]

    def read_bytes(self):
        nbytes = self.read_int()
        assert (nbytes >= 0), nbytes
        return self.read(nbytes)

    def skip_int(self):
        self.skip(4)

    def skip_long(self):
        self.skip(8)

    def skip_bytes(self):
        nbytes = self.read_int()
        assert (nbytes >= 0), nbytes
        self.skip(nbytes)


class FlinkAvroDatumReader(DatumReader):
    """
    Flink-customized datum reader to parse composite data structure, to fit with changed basic type
    serialization in FlinkAvroDecoder.
    """

    def __init__(self, writer_schema=None, reader_schema=None):
        super().__init__(writer_schema, reader_schema)

    def read_array(self, writer_schema, reader_schema, decoder: 'FlinkAvroDecoder'):
        read_items = []
        block_count = decoder.read_var_long()
        assert block_count >= 0
        if block_count == 0:
            return read_items
        for i in range(block_count):
            read_items.append(self.read_data(writer_schema.items,
                                             reader_schema.items, decoder))
        decoder.read_var_long()
        return read_items

    def skip_array(self, writer_schema, decoder: 'FlinkAvroDecoder'):
        block_count = decoder.read_var_long()
        assert block_count >= 0
        if block_count == 0:
            return
        for i in range(block_count):
            self.skip_data(writer_schema.items, decoder)
        decoder.read_var_long()

    def read_map(self, writer_schema, reader_schema, decoder: 'FlinkAvroDecoder'):
        read_items = {}
        block_count = decoder.read_var_long()
        assert block_count >= 0
        if block_count == 0:
            return read_items
        for i in range(block_count):
            key = decoder.read_utf8()
            read_items[key] = self.read_data(writer_schema.values,
                                             reader_schema.values, decoder)
        decoder.read_var_long()
        return read_items

    def skip_map(self, writer_schema, decoder: 'FlinkAvroDecoder'):
        block_count = decoder.read_var_long()
        assert block_count >= 0
        if block_count == 0:
            return
        for i in range(block_count):
            decoder.skip_utf8()
            self.skip_data(writer_schema.values, decoder)
        decoder.read_long()

    def read_union(self, writer_schema, reader_schema, decoder: 'FlinkAvroDecoder'):
        index_of_schema = int(decoder.read_int())
        if index_of_schema >= len(writer_schema.schemas):
            fail_msg = "Can't access branch index %d for union with %d branches" \
                       % (index_of_schema, len(writer_schema.schemas))
            raise SchemaResolutionException(fail_msg, writer_schema, reader_schema)
        selected_writer_schema = writer_schema.schemas[index_of_schema]

        return self.read_data(selected_writer_schema, reader_schema, decoder)

    def skip_union(self, writer_schema, decoder):
        index_of_schema = int(decoder.read_int())
        if index_of_schema >= len(writer_schema.schemas):
            fail_msg = "Can't access branch index %d for union with %d branches" \
                       % (index_of_schema, len(writer_schema.schemas))
            raise SchemaResolutionException(fail_msg, writer_schema)
        return self.skip_data(writer_schema.schemas[index_of_schema], decoder)


class FlinkAvroEncoder(BinaryEncoder):

    def __init__(self, writer):
        super().__init__(writer)

    def write_int(self, datum):
        self.write(STRUCT_INT.pack(datum))

    def write_long(self, datum):
        self.write(STRUCT_LONG_LONG.pack(datum))

    def write_var_long(self, datum):
        while datum & 0x80 != 0:
            self.write((datum & 0x80).to_bytes(1, 'big'))
            datum <<= 7
        self.write(datum.to_bytes(1, 'big'))

    def write_float(self, datum):
        self.write(STRUCT_FLOAT.pack(datum))

    def write_double(self, datum):
        self.write(STRUCT_DOUBLE.pack(datum))

    def write_bytes(self, datum):
        self.write_int(len(datum))
        self.write(datum)


class FlinkAvroDatumWriter(DatumWriter):

    def __init__(self, writer_schema=None):
        super().__init__(writers_schema=writer_schema)

    def write_array(self, writer_schema, datum, encoder):
        if len(datum) > 0:
            encoder.write_var_long(len(datum))
            for item in datum:
                self.write_data(writer_schema.items, item, encoder)
        encoder.write_long(0)

    def write_map(self, writer_schema, datum, encoder):
        if len(datum) > 0:
            encoder.write_var_long(len(datum))
            for key, val in datum.items():
                encoder.write_utf8(key)
                self.write_data(writer_schema.values, val, encoder)
        encoder.write_var_long(0)

    def write_union(self, writer_schema, datum, encoder):
        # resolve union
        index_of_schema = -1
        for i, candidate_schema in enumerate(writer_schema.schemas):
            if validate(candidate_schema, datum):
                index_of_schema = i
        if index_of_schema < 0:
            raise AvroTypeException(writer_schema, datum)

        # write data
        encoder.write_int(index_of_schema)
        self.write_data(writer_schema.schemas[index_of_schema], datum, encoder)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import logging
import sys

import pandas as pd
import numpy as np

from pyflink.table import (DataTypes, TableEnvironment, EnvironmentSettings)


def conversion_from_dataframe():
    t_env = TableEnvironment.create(EnvironmentSettings.in_streaming_mode())
    t_env.get_config().set("parallelism.default", "1")

    # define the source with watermark definition
    pdf = pd.DataFrame(np.random.rand(1000, 2))
    table = t_env.from_pandas(
        pdf,
        schema=DataTypes.ROW([DataTypes.FIELD("a", DataTypes.DOUBLE()),
                              DataTypes.FIELD("b", DataTypes.DOUBLE())]))

    print(table.to_pandas())


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    conversion_from_dataframe()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.java_gateway import get_gateway

__all__ = ['Duration', 'Instant', 'Time']


class Duration(object):
    """
    A time-based amount of time, such as '34.5 seconds'.
    """
    def __init__(self, j_duration):
        self._j_duration = j_duration

    @staticmethod
    def of_days(days: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofDays(days))

    @staticmethod
    def of_hours(hours: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofHours(hours))

    @staticmethod
    def of_millis(millis: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofMillis(millis))

    @staticmethod
    def of_minutes(minutes: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofMinutes(minutes))

    @staticmethod
    def of_nanos(nanos: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofNanos(nanos))

    @staticmethod
    def of_seconds(seconds: int):
        return Duration(get_gateway().jvm.java.time.Duration.ofSeconds(seconds))

    def __eq__(self, other):
        return isinstance(other, Duration) and self._j_duration.equals(other._j_duration)


class Instant(object):
    """
    An instantaneous point on the time-line. Similar to Java.time.Instant.
    """

    def __init__(self, seconds, nanos):
        self.seconds = seconds
        self.nanos = nanos

    def to_epoch_milli(self):
        if self.seconds < 0 < self.nanos:
            return (self.seconds + 1) * 1000 + self.nanos // 1000_1000 - 1000
        else:
            return self.seconds * 1000 + self.nanos // 1000_000

    @staticmethod
    def of_epoch_milli(epoch_milli: int) -> 'Instant':
        secs = epoch_milli // 1000
        mos = epoch_milli % 1000
        return Instant(secs, mos * 1000_000)

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.seconds == other.seconds and
                self.nanos == other.nanos)

    def __repr__(self):
        return 'Instant<{}, {}>'.format(self.seconds, self.nanos)


class Time(object):
    """
    The definition of a time interval.
    """

    def __init__(self, milliseconds: int):
        self._milliseconds = milliseconds

    def to_milliseconds(self) -> int:
        return self._milliseconds

    @staticmethod
    def milliseconds(milliseconds: int):
        return Time(milliseconds)

    @staticmethod
    def seconds(seconds: int):
        return Time.milliseconds(seconds * 1000)

    @staticmethod
    def minutes(minutes: int):
        return Time.seconds(minutes * 60)

    @staticmethod
    def hours(hours: int):
        return Time.minutes(hours * 60)

    @staticmethod
    def days(days: int):
        return Time.hours(days * 24)

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self._milliseconds == other._milliseconds)

    def __str__(self):
        return "{} ms".format(self._milliseconds)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import logging
import sys

from pyflink.common.time import Instant

from pyflink.common import Types
from pyflink.datastream import StreamExecutionEnvironment
from pyflink.table import (DataTypes, TableDescriptor, Schema, StreamTableEnvironment)
from pyflink.table.expressions import lit, col
from pyflink.table.window import Tumble


def tumble_window_demo():
    env = StreamExecutionEnvironment.get_execution_environment()
    env.set_parallelism(1)
    t_env = StreamTableEnvironment.create(stream_execution_environment=env)

    # define the source with watermark definition
    ds = env.from_collection(
        collection=[
            (Instant.of_epoch_milli(1000), 'Alice', 110.1),
            (Instant.of_epoch_milli(4000), 'Bob', 30.2),
            (Instant.of_epoch_milli(3000), 'Alice', 20.0),
            (Instant.of_epoch_milli(2000), 'Bob', 53.1),
            (Instant.of_epoch_milli(5000), 'Alice', 13.1),
            (Instant.of_epoch_milli(3000), 'Bob', 3.1),
            (Instant.of_epoch_milli(7000), 'Bob', 16.1),
            (Instant.of_epoch_milli(10000), 'Alice', 20.1)
        ],
        type_info=Types.ROW([Types.INSTANT(), Types.STRING(), Types.FLOAT()]))

    table = t_env.from_data_stream(
        ds,
        Schema.new_builder()
              .column_by_expression("ts", "CAST(f0 AS TIMESTAMP(3))")
              .column("f1", DataTypes.STRING())
              .column("f2", DataTypes.FLOAT())
              .watermark("ts", "ts - INTERVAL '3' SECOND")
              .build()
    ).alias("ts", "name", "price")

    # define the sink
    t_env.create_temporary_table(
        'sink',
        TableDescriptor.for_connector('print')
                       .schema(Schema.new_builder()
                               .column('name', DataTypes.STRING())
                               .column('total_price', DataTypes.FLOAT())
                               .column('w_start', DataTypes.TIMESTAMP_LTZ())
                               .column('w_end', DataTypes.TIMESTAMP_LTZ())
                               .build())
                       .build())

    # define the tumble window operation
    table = table.window(Tumble.over(lit(5).seconds).on(col("ts")).alias("w")) \
                 .group_by(col('name'), col('w')) \
                 .select(col('name'), col('price').sum, col("w").start, col("w").end)

    # submit for execution
    table.execute_insert('sink') \
         .wait()
    # remove .wait if submitting to a remote cluster, refer to
    # https://nightlies.apache.org/flink/flink-docs-stable/docs/dev/python/faq/#wait-for-jobs-to-finish-when-executing-jobs-in-mini-cluster
    # for more details


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    tumble_window_demo()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC
from typing import cast

from pyflink.datastream import TimerService, TimeDomain
from pyflink.datastream.functions import KeyedProcessFunction, KeyedCoProcessFunction, \
    ProcessFunction, CoProcessFunction, BroadcastProcessFunction, KeyedBroadcastProcessFunction, \
    BaseBroadcastProcessFunction
from pyflink.datastream.state import MapStateDescriptor, BroadcastState, ReadOnlyBroadcastState, \
    OperatorStateStore
from pyflink.fn_execution.internal_state import InternalBroadcastState


class InternalKeyedProcessFunctionOnTimerContext(
        KeyedProcessFunction.OnTimerContext, KeyedCoProcessFunction.OnTimerContext):
    """
    Internal implementation of OnTimerContext of KeyedProcessFunction and KeyedCoProcessFunction.
    """

    def __init__(self, timer_service: TimerService):
        self._timer_service = timer_service
        self._time_domain = None
        self._timestamp = None
        self._current_key = None

    def get_current_key(self):
        return self._current_key

    def set_current_key(self, current_key):
        self._current_key = current_key

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._timestamp

    def set_timestamp(self, ts: int):
        self._timestamp = ts

    def time_domain(self) -> TimeDomain:
        return self._time_domain

    def set_time_domain(self, td: TimeDomain):
        self._time_domain = td


class InternalKeyedProcessFunctionContext(
        KeyedProcessFunction.Context, KeyedCoProcessFunction.Context):
    """
    Internal implementation of Context of KeyedProcessFunction and KeyedCoProcessFunction.
    """

    def __init__(self, timer_service: TimerService):
        self._timer_service = timer_service
        self._timestamp = None
        self._current_key = None

    def get_current_key(self):
        return self._current_key

    def set_current_key(self, current_key):
        self._current_key = current_key

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._timestamp

    def set_timestamp(self, ts: int):
        self._timestamp = ts


class InternalProcessFunctionContext(ProcessFunction.Context, CoProcessFunction.Context):
    """
    Internal implementation of ProcessFunction.Context and CoProcessFunction.Context.
    """

    def __init__(self, timer_service: TimerService):
        self._timer_service = timer_service
        self._timestamp = None

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._timestamp

    def set_timestamp(self, ts: int):
        self._timestamp = ts


class InternalBaseBroadcastProcessFunctionContext(BaseBroadcastProcessFunction.Context, ABC):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        self._timer_service = timer_service
        self._timestamp = None
        self._operator_state_store = operator_state_store

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._timestamp

    def set_timestamp(self, ts: int):
        self._timestamp = ts

    def current_processing_time(self) -> int:
        return self._timer_service.current_processing_time()

    def current_watermark(self) -> int:
        return self._timer_service.current_watermark()


class InternalBroadcastProcessFunctionContext(InternalBaseBroadcastProcessFunctionContext,
                                              BroadcastProcessFunction.Context):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        InternalBaseBroadcastProcessFunctionContext.__init__(
            self, timer_service, operator_state_store
        )

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> BroadcastState:
        return self._operator_state_store.get_broadcast_state(state_descriptor)


class InternalBroadcastProcessFunctionReadOnlyContext(InternalBaseBroadcastProcessFunctionContext,
                                                      BroadcastProcessFunction.ReadOnlyContext):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        InternalBaseBroadcastProcessFunctionContext.__init__(
            self, timer_service, operator_state_store
        )

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return cast(
            InternalBroadcastState,
            self._operator_state_store.get_broadcast_state(state_descriptor)
        ).to_read_only_broadcast_state()


class InternalKeyedBroadcastProcessFunctionContext(InternalBaseBroadcastProcessFunctionContext,
                                                   KeyedBroadcastProcessFunction.Context):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        InternalBaseBroadcastProcessFunctionContext.__init__(
            self, timer_service, operator_state_store
        )

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> BroadcastState:
        return self._operator_state_store.get_broadcast_state(state_descriptor)


class InternalKeyedBroadcastProcessFunctionReadOnlyContext(
        InternalBaseBroadcastProcessFunctionContext,
        KeyedBroadcastProcessFunction.ReadOnlyContext):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        InternalBaseBroadcastProcessFunctionContext.__init__(
            self, timer_service, operator_state_store
        )
        self._current_key = None

    def set_current_key(self, key):
        self._current_key = key

    def get_current_key(self):
        return self._current_key

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return cast(
            InternalBroadcastState,
            self._operator_state_store.get_broadcast_state(state_descriptor)
        ).to_read_only_broadcast_state()


class InternalKeyedBroadcastProcessFunctionOnTimerContext(
        InternalBaseBroadcastProcessFunctionContext,
        KeyedBroadcastProcessFunction.OnTimerContext):

    def __init__(self, timer_service: TimerService, operator_state_store: OperatorStateStore):
        InternalBaseBroadcastProcessFunctionContext.__init__(
            self, timer_service, operator_state_store
        )
        self._current_key = None
        self._time_domain = None

    def set_current_key(self, key):
        self._current_key = key

    def get_current_key(self):
        return self._current_key

    def set_time_domain(self, time_domain):
        self._time_domain = time_domain

    def time_domain(self):
        return self._time_domain

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return cast(
            InternalBroadcastState,
            self._operator_state_store.get_broadcast_state(state_descriptor)
        ).to_read_only_broadcast_state()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# cython: language_level = 3
# cython: infer_types = True
# cython: profile=True
# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
from libc.stdlib cimport free, malloc
from typing import List, Dict

from pyflink.common import Row
from pyflink.fn_execution.coders import PickleCoder
from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \
    PerKeyStateDataViewStore
from pyflink.table import AggregateFunction, TableAggregateFunction

cdef InternalRow join_row(list left, list right, InternalRowKind row_kind):
    return InternalRow(left.__add__(right), row_kind)

cdef class DistinctViewDescriptor:
    def __cinit__(self, input_extractor, filter_args):
        self.input_extractor = input_extractor
        self.filter_length = len(filter_args)
        self.filter_args = <int*> malloc(self.filter_length * sizeof(int))
        for i in range(self.filter_length):
            self.filter_args[i] = filter_args[i]

    def __dealloc__(self):
        if self.filter_args:
            free(self.filter_args)

cdef class RowKeySelector(object):
    """
    A simple key selector used to extract the current key from the input Row according to the
    group-by field indexes.
    """

    def __cinit__(self, list grouping):
        self.length = len(grouping)
        self.grouping = <size_t*> malloc(self.length * sizeof(size_t))
        for i in range(self.length):
            self.grouping[i] = grouping[i]

    cdef list get_key(self, list data):
        cdef size_t i
        return [data[self.grouping[i]] for i in range(self.length)]

    def __dealloc__(self):
        if self.grouping:
            free(self.grouping)

cdef class AggsHandleFunctionBase:
    """
    The base class for handling aggregate or table aggregate functions.
    """
    cdef void open(self, object state_data_view_store):
        """
        Initialization method for the function. It is called before the actual working methods.

        :param state_data_view_store: The object used to manage the DataView.
        """
        pass

    cdef void accumulate(self, InternalRow input_data):
        """
        Accumulates the input values to the accumulators.

        :param input_data: Input values bundled in a InternalRow.
        """
        pass

    cdef void retract(self, InternalRow input_data):
        """
        Retracts the input values from the accumulators.

        :param input_data: Input values bundled in a InternalRow.
        """
        pass

    cdef void merge(self, list accumulators):
        """
        Merges the other accumulators into current accumulators.

        :param accumulators: The other List of accumulators.
        """
        pass

    cdef void set_accumulators(self, list accumulators):
        """
        Set the current accumulators (saved in a List) which contains the current aggregated
        results.

        In streaming: accumulators are stored in the state, we need to restore aggregate buffers
        from state.

        In batch: accumulators are stored in the dict, we need to restore aggregate buffers from
        dict.

        :param accumulators: Current accumulators.
        """
        pass

    cdef list get_accumulators(self):
        """
        Gets the current accumulators (saved in a list) which contains the current
        aggregated results.

        :return: The current accumulators.
        """
        pass

    cdef list create_accumulators(self):
        """
        Initializes the accumulators and save them to an accumulators List.

        :return: A List of accumulators which contains the aggregated results.
        """
        pass

    cdef void cleanup(self):
        """
        Cleanup for the retired accumulators state.
        """
        pass

    cdef void close(self):
        """
        Tear-down method for this function. It can be used for clean up work.
        By default, this method does nothing.
        """
        pass

    cdef list get_value(self):
        """
        Gets the result of the aggregation from the current accumulators.

        :return: The final result (saved in a List) of the current accumulators.
        """
        pass

    cdef list emit_value(self, list current_key, bint is_retract):
        """
        Emit the result of the table aggregation.
        """
        pass

cdef class SimpleAggsHandleFunctionBase(AggsHandleFunctionBase):
    """
    A simple AggsHandleFunction implementation which provides the basic functionality.
    """

    def __init__(self,
                 udfs: List[AggregateFunction],
                 input_extractors: List,
                 udf_data_view_specs: List[List[DataViewSpec]],
                 filter_args: List[int],
                 distinct_indexes: List[int],
                 distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
        self._udfs = udfs
        self._udf_num = len(self._udfs)
        self._input_extractors = input_extractors
        self._accumulators = None
        self._udf_data_view_specs = udf_data_view_specs
        self._udf_data_views = []
        self._filter_args = <int*> malloc(self._udf_num * sizeof(int))
        self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))
        for i in range(self._udf_num):
            self._filter_args[i] = filter_args[i]
            self._distinct_indexes[i] = distinct_indexes[i]
        self._distinct_view_descriptors = distinct_view_descriptors
        self._distinct_data_views = {}

    cdef void open(self, object state_data_view_store):
        cdef dict data_views
        for udf in self._udfs:
            udf.open(state_data_view_store.get_runtime_context())
        for data_view_specs in self._udf_data_view_specs:
            data_views = {}
            for data_view_spec in data_view_specs:
                if isinstance(data_view_spec, ListViewSpec):
                    data_views[data_view_spec.field_index] = \
                        state_data_view_store.get_state_list_view(
                            data_view_spec.state_id,
                            data_view_spec.element_coder)
                elif isinstance(data_view_spec, MapViewSpec):
                    data_views[data_view_spec.field_index] = \
                        state_data_view_store.get_state_map_view(
                            data_view_spec.state_id,
                            data_view_spec.key_coder,
                            data_view_spec.value_coder)
            self._udf_data_views.append(data_views)
        for key in self._distinct_view_descriptors.keys():
            self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
                "agg%ddistinct" % key,
                PickleCoder(),
                PickleCoder())

    cdef void accumulate(self, InternalRow input_data):
        cdef size_t i, j, filter_length
        cdef int distinct_index, filter_arg
        cdef int*filter_args
        cdef bint filtered
        cdef DistinctViewDescriptor distinct_view_descriptor
        cdef object distinct_data_view
        cdef InternalRow internal_row
        for i in range(self._udf_num):
            if i in self._distinct_data_views:
                distinct_view_descriptor = self._distinct_view_descriptors[i]
                filter_length = distinct_view_descriptor.filter_length
                if filter_length == 0:
                    filtered = False
                else:
                    filtered = True
                    filter_args = distinct_view_descriptor.filter_args
                    for j in range(filter_length):
                        filter_arg = filter_args[j]
                        if input_data[filter_arg]:
                            filtered = False
                            break
                if not filtered:
                    args = distinct_view_descriptor.input_extractor(input_data)
                    distinct_data_view = self._distinct_data_views[i]
                    if args in distinct_data_view:
                        distinct_data_view[args] += 1
                    else:
                        distinct_data_view[args] = 1
            filter_arg = self._filter_args[i]
            if filter_arg >= 0 and not input_data[filter_arg]:
                continue
            args = self._input_extractors[i](input_data)
            distinct_index = self._distinct_indexes[i]
            if distinct_index >= 0:
                distinct_data_view = self._distinct_data_views[distinct_index]
                if args in distinct_data_view:
                    if distinct_data_view[args] > 1:
                        continue
                else:
                    raise Exception(
                        "The args are not in the distinct data view, this should not happen.")
            # Convert InternalRow to Row
            if len(args) == 1 and isinstance(args[0], InternalRow):
                internal_row = <InternalRow> args[0]
                args[0] = internal_row.to_row()
            self._udfs[i].accumulate(self._accumulators[i], *args)

    cdef void retract(self, InternalRow input_data):
        cdef size_t i, j, filter_length
        cdef int distinct_index, filter_arg
        cdef bint filtered
        cdef DistinctViewDescriptor distinct_view_descriptor
        cdef object distinct_data_view
        cdef InternalRow internal_row
        for i in range(self._udf_num):
            if i in self._distinct_data_views:
                distinct_view_descriptor = self._distinct_view_descriptors[i]
                filter_length = distinct_view_descriptor.filter_length
                if filter_length == 0:
                    filtered = False
                else:
                    filtered = True
                    for j in range(filter_length):
                        filter_arg = distinct_view_descriptor.filter_args[j]
                        if input_data[filter_arg]:
                            filtered = False
                            break
                if not filtered:
                    args = distinct_view_descriptor.input_extractor(input_data)
                    distinct_data_view = self._distinct_data_views[i]
                    if args in distinct_data_view:
                        distinct_data_view[args] -= 1
                        if distinct_data_view[args] == 0:
                            del distinct_data_view[args]
            filter_arg = self._filter_args[i]
            if filter_arg >= 0 and not input_data[filter_arg]:
                continue
            args = self._input_extractors[i](input_data)
            distinct_index = self._distinct_indexes[i]
            if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:
                continue
            # Convert InternalRow to Row
            if len(args) == 1 and isinstance(args[0], InternalRow):
                internal_row = <InternalRow> args[0]
                args[0] = internal_row.to_row()
            self._udfs[i].retract(self._accumulators[i], *args)

    cdef void merge(self, list accumulators):
        cdef size_t i
        for i in range(self._udf_num):
            self._udfs[i].merge(self._accumulators[i], [accumulators[i]])

    cdef void set_accumulators(self, list accumulators):
        cdef size_t i, index
        if self._udf_data_views:
            for i in range(len(self._udf_data_views)):
                for index, data_view in self._udf_data_views[i].items():
                    accumulators[i][index] = data_view
        self._accumulators = accumulators

    cdef list get_accumulators(self):
        return self._accumulators

    cdef list create_accumulators(self):
        return [udf.create_accumulator() for udf in self._udfs]

    cdef void cleanup(self):
        cdef size_t i
        for i in range(len(self._udf_data_views)):
            for data_view in self._udf_data_views[i].values():
                data_view.clear()

    cdef void close(self):
        for udf in self._udfs:
            udf.close()

    def __dealloc__(self):
        if self._filter_args:
            free(self._filter_args)
        if self._distinct_indexes:
            free(self._distinct_indexes)

cdef class SimpleAggsHandleFunction(SimpleAggsHandleFunctionBase):
    """
    A simple AggsHandleFunction implementation which provides the basic functionality.
    """

    def __init__(self,
                 udfs: List[AggregateFunction],
                 input_extractors: List,
                 index_of_count_star: int,
                 count_star_inserted: bool,
                 udf_data_view_specs: List[List[DataViewSpec]],
                 filter_args: List[int],
                 distinct_indexes: List[int],
                 distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
        super(SimpleAggsHandleFunction, self).__init__(
            udfs, input_extractors, udf_data_view_specs, filter_args, distinct_indexes,
            distinct_view_descriptors)
        temp = [i for i in range(len(udfs))]
        if index_of_count_star >= 0 and count_star_inserted:
            # The record count is used internally, should be ignored by the get_value method.
            temp.remove(index_of_count_star)
        self._get_value_indexes_length = len(temp)
        self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)
        for i in range(self._get_value_indexes_length):
            self._get_value_indexes[i] = temp[i]

    cdef list get_value(self):
        cdef size_t i
        cdef size_t*get_value_indexes
        get_value_indexes = self._get_value_indexes
        return [self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
                for i in range(self._get_value_indexes_length)]

cdef class SimpleTableAggsHandleFunction(SimpleAggsHandleFunctionBase):
    """
    A simple TableAggsHandleFunction implementation which provides the basic functionality.
    """

    def __init__(self,
                 udfs: List[TableAggregateFunction],
                 input_extractors: List,
                 udf_data_view_specs: List[List[DataViewSpec]],
                 filter_args: List[int],
                 distinct_indexes: List[int],
                 distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
        super(SimpleTableAggsHandleFunction, self).__init__(
            udfs, input_extractors, udf_data_view_specs, filter_args, distinct_indexes,
            distinct_view_descriptors)

    cdef list emit_value(self, list current_key, bint is_retract):
        cdef InternalRow result
        cdef list results
        udf = self._udfs[0]  # type: TableAggregateFunction
        results = []
        for x in udf.emit_value(self._accumulators[0]):
            if is_retract:
                result = join_row(current_key, self._convert_to_row(x), InternalRowKind.DELETE)
            else:
                result = join_row(current_key, self._convert_to_row(x), InternalRowKind.INSERT)
            results.append(result)
        return results

    cdef list _convert_to_row(self, data):
        if isinstance(data, Row):
            return data._values
        elif isinstance(data, tuple):
            return list(data)
        else:
            return [data]

cdef class RecordCounter:
    """
    The RecordCounter is used to count the number of input records under the current key.
    """

    cdef bint record_count_is_zero(self, list acc):
        pass

    @staticmethod
    cdef RecordCounter of(int index_of_count_star):
        if index_of_count_star >= 0:
            return RetractionRecordCounter(index_of_count_star)
        else:
            return AccumulationRecordCounter()

cdef class AccumulationRecordCounter(RecordCounter):
    cdef bint record_count_is_zero(self, list acc):
        # when all the inputs are accumulations, the count will never be zero
        return acc is None

cdef class RetractionRecordCounter(RecordCounter):
    def __cinit__(self, int index_of_count_star):
        self._index_of_count_star = index_of_count_star

    cdef bint record_count_is_zero(self, list acc):
        # We store the counter in the accumulator and the counter is never be null
        return acc is None or acc[self._index_of_count_star][0] == 0

cdef class GroupAggFunctionBase:
    def __init__(self,
                 aggs_handle: AggsHandleFunctionBase,
                 key_selector: RowKeySelector,
                 state_backend,
                 state_value_coder,
                 generate_update_before: bool,
                 state_cleaning_enabled: bool,
                 index_of_count_star: int):
        self.aggs_handle = aggs_handle
        self.generate_update_before = generate_update_before
        self.state_cleaning_enabled = state_cleaning_enabled
        self.key_selector = key_selector
        self.state_value_coder = state_value_coder
        self.state_backend = state_backend
        self.record_counter = RecordCounter.of(index_of_count_star)
        self.buffer = {}

    cpdef void open(self, object function_context):
        self.aggs_handle.open(PerKeyStateDataViewStore(function_context, self.state_backend))

    cpdef void close(self):
        self.aggs_handle.close()

    cpdef void on_timer(self, InternalRow key):
        if self.state_cleaning_enabled:
            self.state_backend.set_current_key(list(key.values))
            accumulator_state = self.state_backend.get_value_state(
                "accumulators", self.state_value_coder)
            accumulator_state.clear()
            self.aggs_handle.cleanup()

    cpdef void process_element(self, InternalRow input_data):
        cdef list input_value, key
        input_value = input_data.values
        key = self.key_selector.get_key(input_value)
        try:
            self.buffer[tuple(key)].append(input_data)
        except KeyError:
            self.buffer[tuple(key)] = [input_data]

    cpdef list finish_bundle(self):
        pass

cdef class GroupAggFunction(GroupAggFunctionBase):
    def __init__(self,
                 aggs_handle,
                 key_selector: RowKeySelector,
                 state_backend,
                 state_value_coder,
                 generate_update_before: bool,
                 state_cleaning_enabled: bool,
                 index_of_count_star: int):
        super(GroupAggFunction, self).__init__(
            aggs_handle, key_selector, state_backend, state_value_coder, generate_update_before,
            state_cleaning_enabled, index_of_count_star)

    cpdef list finish_bundle(self):
        cdef list results = []
        cdef bint first_row
        cdef list key, pre_agg_value, new_agg_value, accumulators, input_value, input_rows
        cdef InternalRow retract_row, result_row
        cdef SimpleAggsHandleFunction aggs_handle
        cdef InternalRowKind input_row_kind
        cdef tuple current_key
        cdef size_t input_rows_num, start_index, i
        cdef InternalRow input_data
        cdef object accumulator_state, state_backend
        aggs_handle = <SimpleAggsHandleFunction> self.aggs_handle
        state_backend = self.state_backend
        for current_key in self.buffer:
            input_rows = self.buffer[current_key]
            input_rows_num = len(input_rows)
            key = list(current_key)
            first_row = False
            state_backend.set_current_key(key)
            state_backend.clear_cached_iterators()
            accumulator_state = state_backend.get_value_state(
                "accumulators", self.state_value_coder)
            accumulators = accumulator_state.value()
            start_index = 0
            if accumulators is None:
                for i in range(input_rows_num):
                    input_data = input_rows[i]
                    if input_data.is_retract_msg():
                        start_index += 1
                    else:
                        break
                if start_index == input_rows_num:
                    continue
                accumulators = aggs_handle.create_accumulators()
                first_row = True

            # set accumulators to handler first
            aggs_handle.set_accumulators(accumulators)
            # get previous aggregate result
            pre_agg_value = aggs_handle.get_value()

            for i in range(start_index, input_rows_num):
                input_data = input_rows[i]
                # update aggregate result and set to the newRow
                if input_data.is_accumulate_msg():
                    # accumulate input
                    aggs_handle.accumulate(input_data)
                else:
                    # retract input
                    aggs_handle.retract(input_data)

            # get current aggregate result
            new_agg_value = aggs_handle.get_value()

            # get accumulator
            accumulators = aggs_handle.get_accumulators()

            if not self.record_counter.record_count_is_zero(accumulators):
                # we aggregated at least one record for this key

                # update the state
                accumulator_state.update(accumulators)

                # if this was not the first row and we have to emit retractions
                if not first_row:
                    if pre_agg_value != new_agg_value:
                        # retract previous result
                        if self.generate_update_before:
                            # prepare UPDATE_BEFORE message for previous row
                            retract_row = join_row(key, pre_agg_value,
                                                   InternalRowKind.UPDATE_BEFORE)
                            results.append(retract_row)
                        # prepare UPDATE_AFTER message for new row
                        result_row = join_row(key, new_agg_value, InternalRowKind.UPDATE_AFTER)
                        results.append(result_row)
                else:
                    # this is the first, output new result
                    # prepare INSERT message for new row
                    result_row = join_row(key, new_agg_value, InternalRowKind.INSERT)
                    results.append(result_row)
            else:
                # we retracted the last record for this key
                # sent out a delete message
                if not first_row:
                    # prepare delete message for previous row
                    result_row = join_row(key, pre_agg_value, InternalRowKind.DELETE)
                    results.append(result_row)
                # and clear all state
                accumulator_state.clear()
                # cleanup dataview under current key
                aggs_handle.cleanup()
        self.buffer = {}
        return results

cdef class GroupTableAggFunction(GroupAggFunctionBase):
    def __init__(self,
                 aggs_handle,
                 key_selector: RowKeySelector,
                 state_backend,
                 state_value_coder,
                 generate_update_before: bool,
                 state_cleaning_enabled: bool,
                 index_of_count_star: int):
        super(GroupTableAggFunction, self).__init__(
            aggs_handle, key_selector, state_backend, state_value_coder, generate_update_before,
            state_cleaning_enabled, index_of_count_star)

    cpdef list finish_bundle(self):
        cdef bint first_row
        cdef list key, accumulators, input_value, results
        cdef SimpleTableAggsHandleFunction aggs_handle
        cdef InternalRowKind input_row_kind
        cdef tuple current_key
        cdef InternalRow input_data
        cdef size_t start_index, i, input_rows_num
        cdef object state_backend, accumulator_state
        results = []
        aggs_handle = <SimpleTableAggsHandleFunction> self.aggs_handle
        state_backend = self.state_backend
        for current_key in self.buffer:
            input_rows = self.buffer[current_key]
            input_rows_num = len(input_rows)
            key = list(current_key)
            first_row = False
            state_backend.set_current_key(key)
            state_backend.clear_cached_iterators()
            accumulator_state = state_backend.get_value_state(
                "accumulators", self.state_value_coder)
            accumulators = accumulator_state.value()
            start_index = 0
            if accumulators is None:
                for i in range(input_rows_num):
                    input_data = input_rows[i]
                    if input_data.is_retract_msg():
                        start_index += 1
                    else:
                        break
                if start_index == input_rows_num:
                    continue
                accumulators = aggs_handle.create_accumulators()
                first_row = True

            # set accumulators to handler first
            aggs_handle.set_accumulators(accumulators)

            if not first_row and self.generate_update_before:
                results.extend(aggs_handle.emit_value(key, True))

            for i in range(start_index, input_rows_num):
                input_data = input_rows[i]
                # update aggregate result and set to the newRow
                if input_data.is_accumulate_msg():
                    # accumulate input
                    aggs_handle.accumulate(input_data)
                else:
                    # retract input
                    aggs_handle.retract(input_data)

            # get accumulator
            accumulators = aggs_handle.get_accumulators()

            if not self.record_counter.record_count_is_zero(accumulators):
                results.extend(aggs_handle.emit_value(key, False))
                accumulator_state.update(accumulators)
            else:
                # and clear all state
                accumulator_state.clear()
                # cleanup dataview under current key
                aggs_handle.cleanup()
        self.buffer = {}
        return results



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from enum import Enum

from pyflink.java_gateway import get_gateway

__all__ = ['JobStatus']


class JobStatus(Enum):
    """
    Possible states of a job once it has been accepted by the job manager.

    :data:`CREATED`:

    Job is newly created, no task has started to run.

    :data:`RUNNING`:

    Some tasks are scheduled or running, some may be pending, some may be finished.

    :data:`FAILING`:

    The job has failed and is currently waiting for the cleanup to complete.

    :data:`FAILED`:

    The job has failed with a non-recoverable task failure.

    :data:`CANCELLING`:

    Job is being cancelled.

    :data:`CANCELED`:

    Job has been cancelled.

    :data:`FINISHED`:

    All of the job's tasks have successfully finished.

    :data:`RESTARTING`:

    The job is currently undergoing a reset and total restart.

    :data:`SUSPENDED`:

    The job has been suspended which means that it has been stopped but not been removed from a
    potential HA job store.

    :data:`RECONCILING`:

    The job is currently reconciling and waits for task execution report to recover state.

    .. versionadded:: 1.11.0
    """

    CREATED = 0
    RUNNING = 1
    FAILING = 2
    FAILED = 3
    CANCELLING = 4
    CANCELED = 5
    FINISHED = 6
    RESTARTING = 7
    SUSPENDED = 8
    RECONCILING = 9

    def is_globally_terminal_state(self) -> bool:
        """
        Checks whether this state is <i>globally terminal</i>. A globally terminal job
        is complete and cannot fail any more and will not be restarted or recovered by another
        standby master node.

        When a globally terminal state has been reached, all recovery data for the job is
        dropped from the high-availability services.

        :return: ``True`` if this job status is globally terminal, ``False`` otherwise.

        .. versionadded:: 1.11.0
        """
        return self._to_j_job_status().isGloballyTerminalState()

    def is_terminal_state(self) -> bool:
        """
        Checks whether this state is locally terminal. Locally terminal refers to the
        state of a job's execution graph within an executing JobManager. If the execution graph
        is locally terminal, the JobManager will not continue executing or recovering the job.

        The only state that is locally terminal, but not globally terminal is SUSPENDED,
        which is typically entered when the executing JobManager loses its leader status.

        :return: ``True`` if this job status is terminal, ``False`` otherwise.

        .. versionadded:: 1.11.0
        """
        return self._to_j_job_status().isTerminalState()

    @staticmethod
    def _from_j_job_status(j_job_status) -> 'JobStatus':
        return JobStatus[j_job_status.name()]

    def _to_j_job_status(self):
        gateway = get_gateway()
        JJobStatus = gateway.jvm.org.apache.flink.api.common.JobStatus
        return getattr(JJobStatus, self.name)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common.completable_future import CompletableFuture
from pyflink.common.job_execution_result import JobExecutionResult
from pyflink.common.job_id import JobID
from pyflink.common.job_status import JobStatus

__all__ = ['JobClient']


class JobClient(object):
    """
    A client that is scoped to a specific job.

    .. versionadded:: 1.11.0
    """

    def __init__(self, j_job_client):
        self._j_job_client = j_job_client

    def get_job_id(self) -> JobID:
        """
        Returns the JobID that uniquely identifies the job this client is scoped to.

        :return: JobID, or null if the job has been executed on a runtime without JobIDs
                 or if the execution failed.

        .. versionadded:: 1.11.0
        """
        return JobID(self._j_job_client.getJobID())

    def get_job_status(self) -> CompletableFuture:
        """
        Requests the JobStatus of the associated job.

        :return: A CompletableFuture containing the JobStatus of the associated job.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(self._j_job_client.getJobStatus(), JobStatus._from_j_job_status)

    def cancel(self) -> CompletableFuture:
        """
        Cancels the associated job.

        :return: A CompletableFuture for canceling the associated job.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(self._j_job_client.cancel())

    def stop_with_savepoint(self, advance_to_end_of_event_time: bool,
                            savepoint_directory: str = None) -> CompletableFuture:
        """
        Stops the associated job on Flink cluster.

        Stopping works only for streaming programs. Be aware, that the job might continue to run
        for a while after sending the stop command, because after sources stopped to emit data all
        operators need to finish processing.

        :param advance_to_end_of_event_time: Flag indicating if the source should inject a
                                             MAX_WATERMARK in the pipeline.
        :param savepoint_directory: Directory the savepoint should be written to.
        :return: A CompletableFuture containing the path where the savepoint is located.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(
            self._j_job_client.stopWithSavepoint(advance_to_end_of_event_time, savepoint_directory),
            str)

    def trigger_savepoint(self, savepoint_directory: str = None) -> CompletableFuture:
        """
        Triggers a savepoint for the associated job. The savepoint will be written to the given
        savepoint directory.

        :param savepoint_directory: Directory the savepoint should be written to.
        :return: A CompletableFuture containing the path where the savepoint is located.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(self._j_job_client.triggerSavepoint(savepoint_directory), str)

    def get_accumulators(self) -> CompletableFuture:
        """
        Requests the accumulators of the associated job. Accumulators can be requested while it
        is running or after it has finished. The class loader is used to deserialize the incoming
        accumulator results.

        :param class_loader: Class loader used to deserialize the incoming accumulator results.
        :return: A CompletableFuture containing the accumulators of the associated job.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(self._j_job_client.getAccumulators(), dict)

    def get_job_execution_result(self) -> CompletableFuture:
        """
        Returns the JobExecutionResult result of the job execution of the submitted job.

        :return: A CompletableFuture containing the JobExecutionResult result of the job execution.

        .. versionadded:: 1.11.0
        """
        return CompletableFuture(self._j_job_client.getJobExecutionResult(),
                                 JobExecutionResult)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC
from typing import List, Iterable, Tuple, Dict, Collection

from pyflink.datastream import ReduceFunction, AggregateFunction
from pyflink.datastream.state import (T, IN, OUT, V, K, State)
from pyflink.fn_execution.embedded.converters import (DataConverter, DictDataConverter,
                                                      ListDataConverter)
from pyflink.fn_execution.internal_state import (InternalValueState, InternalKvState,
                                                 InternalListState, InternalReducingState,
                                                 InternalAggregatingState, InternalMapState,
                                                 N, InternalReadOnlyBroadcastState,
                                                 InternalBroadcastState)


class StateImpl(State, ABC):
    def __init__(self,
                 state,
                 value_converter: DataConverter):
        self._state = state
        self._value_converter = value_converter

    def clear(self):
        self._state.clear()


class KeyedStateImpl(StateImpl, InternalKvState, ABC):

    def __init__(self,
                 state,
                 value_converter: DataConverter,
                 window_converter: DataConverter = None):
        super(KeyedStateImpl, self).__init__(state, value_converter)
        self._window_converter = window_converter

    def set_current_namespace(self, namespace) -> None:
        j_window = self._window_converter.to_external(namespace)
        self._state.setCurrentNamespace(j_window)


class ValueStateImpl(KeyedStateImpl, InternalValueState):
    def __init__(self,
                 value_state,
                 value_converter: DataConverter,
                 window_converter: DataConverter = None):
        super(ValueStateImpl, self).__init__(value_state, value_converter, window_converter)

    def value(self) -> T:
        return self._value_converter.to_internal(self._state.value())

    def update(self, value: T) -> None:
        self._state.update(self._value_converter.to_external(value))


class ListStateImpl(KeyedStateImpl, InternalListState):

    def __init__(self,
                 list_state,
                 value_converter: ListDataConverter,
                 window_converter: DataConverter = None):
        super(ListStateImpl, self).__init__(list_state, value_converter, window_converter)
        self._element_converter = value_converter._field_converter

    def update(self, values: List[T]) -> None:
        self._state.update(self._value_converter.to_external(values))

    def add_all(self, values: List[T]) -> None:
        self._state.addAll(self._value_converter.to_external(values))

    def get(self) -> OUT:
        states = self._value_converter.to_internal(self._state.get())
        if states:
            yield from states

    def add(self, value: IN) -> None:
        self._state.add(self._element_converter.to_external(value))

    def merge_namespaces(self, target: N, sources: Collection[N]) -> None:
        j_target = self._window_converter.to_external(target)
        j_sources = [self._window_converter.to_external(window) for window in sources]
        self._state.mergeNamespaces(j_target, j_sources)


class ReducingStateImpl(KeyedStateImpl, InternalReducingState):

    def __init__(self,
                 value_state,
                 value_converter: DataConverter,
                 reduce_function: ReduceFunction,
                 window_converter: DataConverter = None):
        super(ReducingStateImpl, self).__init__(value_state, value_converter, window_converter)
        self._reduce_function = reduce_function

    def get(self) -> OUT:
        return self._value_converter.to_internal(self._state.value())

    def add(self, value: IN) -> None:
        if value is None:
            self.clear()
        else:
            current_value = self.get()

            if current_value is None:
                reduce_value = value
            else:
                reduce_value = self._reduce_function.reduce(current_value, value)

            self._state.update(self._value_converter.to_external(reduce_value))

    def merge_namespaces(self, target: N, sources: Collection[N]) -> None:
        merged = None
        for source in sources:
            self.set_current_namespace(source)
            source_state = self.get()

            if source_state is None:
                continue

            self.clear()

            if merged is None:
                merged = source_state
            else:
                merged = self._reduce_function.reduce(merged, source_state)

        if merged is not None:
            self.set_current_namespace(target)
            self._state.update(self._value_converter.to_external(merged))


class AggregatingStateImpl(KeyedStateImpl, InternalAggregatingState):
    def __init__(self,
                 value_state,
                 value_converter,
                 agg_function: AggregateFunction,
                 window_converter: DataConverter = None):
        super(AggregatingStateImpl, self).__init__(value_state, value_converter, window_converter)
        self._agg_function = agg_function

    def get(self) -> OUT:
        accumulator = self._value_converter.to_internal(self._state.value())

        if accumulator is None:
            return None
        else:
            return self._agg_function.get_result(accumulator)

    def add(self, value: IN) -> None:
        if value is None:
            self.clear()
        else:
            accumulator = self._value_converter.to_internal(self._state.value())

            if accumulator is None:
                accumulator = self._agg_function.create_accumulator()

            accumulator = self._agg_function.add(value, accumulator)
            self._state.update(self._value_converter.to_external(accumulator))

    def merge_namespaces(self, target: N, sources: Collection[N]) -> None:
        merged = None
        for source in sources:
            self.set_current_namespace(source)
            source_state = self.get()

            if source_state is None:
                continue

            self.clear()

            if merged is None:
                merged = source_state
            else:
                merged = self._agg_function.merge(merged, source_state)

        if merged is not None:
            self.set_current_namespace(target)
            self._state.update(self._value_converter.to_external(merged))


class MapStateImpl(KeyedStateImpl, InternalMapState):
    def __init__(self,
                 map_state,
                 map_converter: DictDataConverter,
                 window_converter: DataConverter = None):
        super(MapStateImpl, self).__init__(map_state, map_converter, window_converter)
        self._k_converter = map_converter._key_converter
        self._v_converter = map_converter._value_converter

    def get(self, key: K) -> V:
        return self._v_converter.to_internal(
            self._state.get(self._k_converter.to_external(key)))

    def put(self, key: K, value: V) -> None:
        self._state.put(self._k_converter.to_external(key), self._v_converter.to_external(value))

    def put_all(self, dict_value: Dict[K, V]) -> None:
        self._state.putAll(self._value_converter.to_external(dict_value))

    def remove(self, key: K) -> None:
        self._state.remove(self._k_converter.to_external(key))

    def contains(self, key: K) -> bool:
        return self._state.contains(self._k_converter.to_external(key))

    def items(self) -> Iterable[Tuple[K, V]]:
        entries = self._state.entries()
        if entries:
            for entry in entries:
                yield (self._k_converter.to_internal(entry.getKey()),
                       self._v_converter.to_internal(entry.getValue()))

    def keys(self) -> Iterable[K]:
        keys = self._state.keys()
        if keys:
            for k in keys:
                yield self._k_converter.to_internal(k)

    def values(self) -> Iterable[V]:
        values = self._state.values()
        if values:
            for v in values:
                yield self._v_converter.to_internal(v)

    def is_empty(self) -> bool:
        return self._state.isEmpty()


class ReadOnlyBroadcastStateImpl(StateImpl, InternalReadOnlyBroadcastState):

    def __init__(self,
                 map_state,
                 map_converter: DictDataConverter):
        super(ReadOnlyBroadcastStateImpl, self).__init__(map_state, map_converter)
        self._k_converter = map_converter._key_converter
        self._v_converter = map_converter._value_converter

    def get(self, key: K) -> V:
        return self._v_converter.to_internal(
            self._state.get(self._k_converter.to_external(key)))

    def contains(self, key: K) -> bool:
        return self._state.contains(self._k_converter.to_external(key))

    def items(self) -> Iterable[Tuple[K, V]]:
        entries = self._state.entries()
        for entry in entries:
            yield (self._k_converter.to_internal(entry.getKey()),
                   self._v_converter.to_internal(entry.getValue()))

    def keys(self) -> Iterable[K]:
        for k in self._state.keys():
            yield self._k_converter.to_internal(k)

    def values(self) -> Iterable[V]:
        for v in self._state.values():
            yield self._v_converter.to_internal(v)

    def is_empty(self) -> bool:
        return self._state.isEmpty()


class BroadcastStateImpl(ReadOnlyBroadcastStateImpl, InternalBroadcastState):
    def __init__(self,
                 map_state,
                 map_converter: DictDataConverter):
        super(BroadcastStateImpl, self).__init__(map_state, map_converter)
        self._map_converter = map_converter
        self._k_converter = map_converter._key_converter
        self._v_converter = map_converter._value_converter

    def to_read_only_broadcast_state(self) -> InternalReadOnlyBroadcastState[K, V]:
        return ReadOnlyBroadcastStateImpl(self._state, self._map_converter)

    def put(self, key: K, value: V) -> None:
        self._state.put(self._k_converter.to_external(key), self._v_converter.to_external(value))

    def put_all(self, dict_value: Dict[K, V]) -> None:
        self._state.putAll(self._value_converter.to_external(dict_value))

    def remove(self, key: K) -> None:
        self._state.remove(self._k_converter.to_external(key))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import io

from pyflink.serializers import IterableSerializer
from pyflink.table.utils import arrow_to_pandas, pandas_to_arrow


class ArrowSerializer(IterableSerializer):
    """
    Serializes pandas.Series into Arrow streaming format data.
    """

    def __init__(self, schema, row_type, timezone):
        super(ArrowSerializer, self).__init__()
        self._schema = schema
        self._field_types = row_type.field_types()
        self._timezone = timezone

    def __repr__(self):
        return "ArrowSerializer"

    def serialize(self, iterable, stream):
        writer = None
        try:
            for cols in iterable:
                batch = pandas_to_arrow(self._schema, self._timezone, self._field_types, cols)
                if writer is None:
                    import pyarrow as pa
                    writer = pa.RecordBatchStreamWriter(stream, batch.schema)
                writer.write_batch(batch)
        finally:
            if writer is not None:
                writer.close()

    def deserialize(self, stream):
        import pyarrow as pa
        reader = pa.ipc.open_stream(stream)
        for batch in reader:
            yield arrow_to_pandas(self._timezone, self._field_types, [batch])

    def load_from_iterator(self, iter):
        class IteratorIO(io.RawIOBase):
            def __init__(self, iter):
                super(IteratorIO, self).__init__()
                self.iter = iter
                self.leftover = None

            def readable(self):
                return True

            def readinto(self, b):
                output_buffer_len = len(b)
                input = self.leftover or (self.iter.next() if self.iter.hasNext() else None)
                if input is None:
                    return 0
                output, self.leftover = input[:output_buffer_len], input[output_buffer_len:]
                b[:len(output)] = output
                return len(output)
        import pyarrow as pa
        reader = pa.ipc.open_stream(
            io.BufferedReader(IteratorIO(iter), buffer_size=io.DEFAULT_BUFFER_SIZE))
        for batch in reader:
            yield batch
